#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   config_management.py
@Time    :   2024-10-22 20:29:20
@Author  :   SivanLaai
@Version :   1.0
@Contact :   lyhhap@163.com
@Desc    :   
'''
import sys
import os
current_dir = os.path.split(os.path.abspath(__file__))[0]
tool_dir = current_dir.rsplit('/', 2)[0]
sys.path.append(tool_dir)
from enum import Enum
from database.enum import EdaTool, FlowStep 
from flow.flow_db import DbFlow
from workspace.path import WorkspacePath
from flow.flow_manager import FlowManager

from application.DSE.parameter import iEDAParameter, InnovusParameter
import logging

class ConfigManagement:
    def __init__(self, args, eda_tool):
        self.args = args
        self.step = None
        root = self.args.root
        tech = self.args.tech
        design = self.args.project_name
        self.dir_workspace = f"{root}/{tech}/{design}"
        self.workspace = WorkspacePath(self.dir_workspace)
        self.flow_manager = FlowManager(self.workspace.get_config('flow'))

        self.flow_list = list() 
        self.initEdaTool(eda_tool)
        self.initFlowStep()
        self.initConfigPathList()
        self.loadParameters()
        
    def initEdaTool(self, eda_tool):
        if isinstance(eda_tool, Enum):
            self.eda_tool = eda_tool
            return
        if eda_tool == EdaTool.IEDA.value:
            self.eda_tool = EdaTool.IEDA
        elif eda_tool== EdaTool.INNOVUS.value:
            self.eda_tool = EdaTool.INNOVUS
        elif eda_tool==EdaTool.DREAMPLACE.value:
            self.eda_tool = EdaTool.DREAMPLACE
    
    def getEdaTool(self, eda_tool=None):
        if eda_tool is not None:
            if isinstance(eda_tool, Enum):
                return eda_tool
            if eda_tool == EdaTool.IEDA.value:
                eda_tool = EdaTool.IEDA
            elif eda_tool== EdaTool.INNOVUS.value:
                eda_tool = EdaTool.INNOVUS
            elif eda_tool==EdaTool.DREAMPLACE.value:
                eda_tool = EdaTool.DREAMPLACE
            return eda_tool
        return self.eda_tool

    def initFlowList(self):
        if self.step == FlowStep.full_flow:
            tmp_flows = self.args.flows.split(",")
        else:
            tmp_flows = self.args.step.split(",")
        self.flow_manager = FlowManager(self.workspace.get_config('flow'))
        for db_flow in self.flow_manager.flow_list:
            if db_flow.step.value.lower()==tmp_flows[0]:
                self.flow_list.append(DbFlow(eda_tool = self.eda_tool,step = db_flow.step,pre_flow=db_flow.pre_flow))
                break
        for i in range(1, len(tmp_flows)):
            pre_step = self.checkFlowStep(tmp_flows[i-1]) 
            step = self.checkFlowStep(tmp_flows[i]) 
            # print(step)
            self.flow_list.append(DbFlow(eda_tool = self.eda_tool,step = step,pre_flow=self.flow_list[-1]))
        # print(self.flow_list)
        for flow in self.flow_list:
            print(flow.pre_flow.step)
            print(flow.step)
        # exit(0)

    
    def getFlowList(self):
        return self.flow_list
    
    def getFlowStepsByStepArg(self, step_list):
        flow_steps = list()
        for step in step_list:
            flow_steps.append(self.checkFlowStep(step))
        return flow_steps
    
    def checkFlowStep(self, step):
        if step == FlowStep.floorplan.value.lower():
            return FlowStep.floorplan
        
        if step == FlowStep.fixFanout.value.lower():
            return FlowStep.fixFanout
            
        if step == FlowStep.place.value.lower():
            return FlowStep.place
        
        if(step == FlowStep.cts.value.lower()):
            return FlowStep.cts
            
        if step == FlowStep.optDrv.value.lower():
            return FlowStep.optDrv
            
        if step == FlowStep.optHold.value.lower():
            return FlowStep.optHold
            
        if step == FlowStep.optSetup.value.lower():
            return FlowStep.optSetup
            
        if step == FlowStep.legalization.value.lower():
            return FlowStep.legalization
            
        if (step == FlowStep.route.value.lower()):
            return FlowStep.route
            
        if step == FlowStep.filler.value.lower():
            return FlowStep.filler
            
        if step == FlowStep.gds.value.lower():
            return FlowStep.gds
        
        if step == FlowStep.drc.value.lower():
            return FlowStep.drc

    def initFlowStep(self):
        if self.args.step == FlowStep.full_flow.value.lower():
            self.step = FlowStep.full_flow
        elif self.args.step == FlowStep.place.value.lower():
            self.step = FlowStep.place
        elif self.args.step == FlowStep.cts.value.lower():
            self.step = FlowStep.cts
        else:
            pass
        self.initFlowList()
    
    def getConfigPath(self, step):
        config_path = self.workspace.get_config(self.eda_tool.value+"_config").get(step.value, None)
        return config_path

    
    def initConfigPathList(self):
        self.config_paths = dict()
        self.best_config_paths = dict()
        # print("initWorkspace", self.flow_list)
        for db_flow in self.flow_list:
            step = db_flow.step
            step_config_path = self.workspace.get_config(self.eda_tool.value+"_config").get(step.value, None)
            best_step_config_path = self.workspace.get_config(self.eda_tool.value+"_config").get(step.value, None).replace(".json", "_best.json")
            if step_config_path is not None:
                self.config_paths[step.value.lower()] = step_config_path
                self.best_config_paths[step.value.lower()] = best_step_config_path

        if not len(self.config_paths):
            logging.error("self.config_paths is empty, process exit")


    def getConfigPathList(self):
        return self.config_paths
    
    def getBestConfigPathList(self):
        return self.best_config_paths

    def getStep(self):
        return self.step

    def getWorkspacePath(self):
        return  self.dir_workspace

    def loadParameters(self):
        # load parameters
        # self.config_path = self.args.config_path
        # parameter space for iEDA
        eda_tool = self.eda_tool
        if eda_tool==EdaTool.IEDA:
            self.params = iEDAParameter(step=self.step)
            self.params.load_list(self.config_paths)
            # print(self.params.__dict__)
        elif eda_tool==EdaTool.INNOVUS:
            self.params = InnovusParameter()
            self.args.config_path = os.path.join(os.path.dirname(__file__), "config/Innovus/default.json")
            self.config_path = self.args.config_path
            #print(self.config_path)
            self.params.load(self.config_path)
        elif eda_tool==EdaTool.DREAMPLACE:
            pass
    
    def getParameters(self):
        return self.params


#!/usr/bin/python
# -*- encoding: utf-8 -*-
'''
@file         : main.py
@author       : Yell
@brief        : 
@version      : 0.1
@date         : 2024-03-02 09:10:20
'''

import gzip

from utility.json_parser import JsonParser

class ReportJson():
    """basic class for transfer report to json"""
    def __init__(self, report_path : str, json_path : str):
        self.report_path = report_path
        self.json_path = json_path
        self.report_parser = None
        self.json_parser = None
        
    def get_report_parser(self):
        return self.report_parser
    
    def get_json_data(self):
        return self.json_parser.get_json_data()
    
    def read(self):
        # read report file
        with gzip.open(self.report_path, 'rb') as f:
            self.report_parser = f.read()
            
        # read json file
        self.json_parser = JsonParser(self.json_path)
        self.json_parser.read_create()
    
    def writeJson(self):
        return self.json_parser.write()


#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
'''
@File : layge_model.py
@Author : yell
@Desc : large model feature database
'''


from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional


@dataclass
class LmNode:
    id: Optional[int] = None
    x: Optional[int] = None
    y: Optional[int] = None
    real_x: Optional[int] = None
    real_y: Optional[int] = None
    row: Optional[int] = None
    col: Optional[int] = None
    layer: Optional[int] = None
    pin_id: Optional[int] = None


@dataclass
class LmPath:
    node1: Optional[LmNode] = None
    node2: Optional[LmNode] = None


@dataclass
class LmWireFeature:
    wire_width: Optional[int] = None
    wire_len: Optional[int] = None
    drc_num: Optional[int] = None
    R: Optional[float] = None
    C: Optional[float] = None
    power: Optional[float] = None
    delay: Optional[float] = None
    slew: Optional[float] = None
    congestion: Optional[float] = None
    wire_density: Optional[float] = None
    drc_type: List[str] = field(default_factory=list)


@dataclass
class LmWire:
    id: Optional[int] = None
    feature: Optional[LmWireFeature] = None
    wire: Optional[LmPath] = None
    path_num: Optional[int] = None
    paths: List[LmPath] = field(default_factory=list)


@dataclass
class LmPin:
    id: Optional[int] = None
    pin_name: Optional[str] = None
    instance: Optional[str] = None
    is_driver: Optional[str] = None


@dataclass
class LmNetFeature:
    llx: Optional[int] = None
    lly: Optional[int] = None
    urx: Optional[int] = None
    ury: Optional[int] = None
    wire_len: Optional[int] = None
    via_num: Optional[int] = None
    drc_num: Optional[int] = None
    R: Optional[float] = None
    C: Optional[float] = None
    power: Optional[float] = None
    delay: Optional[float] = None
    slew: Optional[float] = None
    fanout: Optional[int] = None
    aspect_ratio: Optional[int] = None
    width: Optional[int] = None
    height: Optional[int] = None
    area: Optional[int] = None
    l_ness: Optional[float] = None
    drc_type: List[str] = field(default_factory=list)
    volume: Optional[int] = None
    layer_ratio: List[float] = field(default_factory=list)
    rsmt: Optional[int] = None


@dataclass
class LmNetRoutingPoint:
    x: int
    y: int
    layer_id: int

    def __str__(self):
        return f"({self.x}, {self.y}, {self.layer_id})"

    def __repr__(self):
        return f"Point({self.x}, {self.y}, {self.layer_id})"

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y and self.layer_id == other.layer_id

    def __hash__(self):
        return hash((self.x, self.y, self.layer_id))


@dataclass
class LmNetRoutingVertex:
    id: int
    is_pin: bool
    is_driver_pin: bool
    point: LmNetRoutingPoint


@dataclass
class LmNetRoutingEdge:
    source_id: int
    target_id: int
    path: List[LmNetRoutingPoint]


@dataclass
class LmNetRoutingGraph:
    vertices: List[LmNetRoutingVertex]
    edges: List[LmNetRoutingEdge]


@dataclass
class LmNet:
    id: Optional[int] = None
    name: Optional[str] = None
    feature: Optional[LmNetFeature] = None
    pin_num: Optional[int] = None
    pins: List[LmPin] = field(default_factory=list)
    wire_num: Optional[int] = None
    wires: List[LmWire] = field(default_factory=list)
    routing_graph: Optional[LmNetRoutingGraph] = None


@dataclass
class LmPatchLayer:
    id: Optional[int] = None
    net_num: Optional[int] = None
    nets: List[LmNet] = field(default_factory=list)
    wire_width: Optional[int] = None
    wire_len: Optional[int] = None
    wire_density: Optional[float] = None
    congestion: Optional[float] = None


@dataclass
class LmPatch:
    id: Optional[int] = None
    patch_id_row: Optional[int] = None
    patch_id_col: Optional[int] = None
    llx: Optional[int] = None
    lly: Optional[int] = None
    urx: Optional[int] = None
    ury: Optional[int] = None
    row_min: Optional[int] = None
    row_max: Optional[int] = None
    col_min: Optional[int] = None
    col_max: Optional[int] = None
    patch_layer: List[LmPatchLayer] = field(default_factory=list)
    area: Optional[int] = None
    cell_density: Optional[float] = None
    pin_density: Optional[int] = None
    net_density: Optional[float] = None
    macro_margin: Optional[int] = None
    RUDY_congestion: Optional[float] = None
    EGR_congestion: Optional[float] = None
    timing_map: Optional[float] = None
    power_map: Optional[float] = None
    ir_drop_map: Optional[float] = None


@dataclass
class LmWirePatternPoint(object):
    x: int = None
    y: int = None
    z: int = None


@dataclass
class LmWirePatternDirection(Enum):
    TOP = "TOP"
    BOTTOM = "BOTTOM"
    LEFT = "LEFT"
    RIGHT = "RIGHT"
    VIA = "VIA"


@dataclass
class LmWirePatternUnit:
    direction: LmWirePatternDirection = None
    length: int = None


@dataclass
class LmWirePatternSeq:
    name: str = None
    units: List[LmWirePatternUnit] = field(default_factory=list)


@dataclass
class LmNetSeq:
    loc_seq: List[LmWirePatternPoint] = field(default_factory=list)
    pattern_seq: List[LmWirePatternSeq] = field(default_factory=list)


@dataclass
class LmTimingWireGraphNode:
    name: str = None
    is_pin: bool = False
    is_port: bool = False


@dataclass
class LmTimingWireGraphEdge:
    from_node: int = None
    to_node: int = None


@dataclass
class LmTimingWireGraph(object):
    nodes: List[LmTimingWireGraphNode] = field(default_factory=list)
    edges: List[LmTimingWireGraphEdge] = field(default_factory=list)


@dataclass
class LmTimingWirePathGraph(object):
    nodes: List[LmTimingWireGraphNode] = field(default_factory=list)
    edges: List[LmTimingWireGraphEdge] = field(default_factory=list)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : folder_permission.py
@Author : yell
@Desc : change folder authorization
'''

######################################################################################
# import ai-eda as root

######################################################################################
import os

class FolderPermissionManager():
    def __init__(self, folder: str):
        self.directory = folder

    def enable_read_and_write(self):
        self.recursive(self.directory)
            
    def recursive(self, directory : str):
        for root, dirs, files in os.walk(directory):
            self.try_chmod(root)
                
            for file in files:
                self.try_chmod(os.path.join(root, file))
                
            for dir in dirs:
                self.try_chmod(os.path.join(root, dir))
                self.recursive(dir)
                
    def try_chmod(self, path):
        try:
            os.chmod(path, 0o777)
        except Exception as e:
            pass
            # print("chmod error for path : " + path)
            
                
                
        




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : gds.py
@Author : yell
@Desc : gds framework
'''
from engine.base import EngineBase
from database.enum import EdaTool, FlowStep, TaskOption
from flow.flow_db import DbFlow
from tools.iEDA.module.gds import IEDAGds

class EngineGDS(EngineBase):
    """gds framework"""
    def __init__(self, dir_workspace : str, dir_resource : str = "",
                 eda_tool : EdaTool = EdaTool.IEDA, task : TaskOption = TaskOption.RUN_EDA,
                 input_def : str = None, input_verilog : str = None, 
                 output_def : str = None, output_verilog : str = None,
                 pre_step : DbFlow = None, output_gds : str=None):
        super().__init__(dir_workspace, dir_resource,
                 eda_tool, task,
                 input_def, input_verilog, 
                 output_def , output_verilog,
                 pre_step)
        self.output_gds = output_gds
    
    def build_path(self):   
        """rebuild path by flow step"""
        # reset input def
        if(self.input_def == None):
            self.input_def =  self.workspace.get_input_def(self.step)
                
        if(self.output_gds == None):
            self.output_gds = self.workspace.get_output_gds(self.step)
        
    def run(self):
        return super().run()
        
    def run_ieda(self):
        # build flow
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                              step=FlowStep.filler)
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                      pre_flow=self.pre_step,
                      step=FlowStep.gds)
        self.build_path()
        
        # run iEDA
        run_flow = IEDAGds(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_gds = self.output_gds)
        run_flow.save_gds()
        
        
        self.result.output_gds=self.output_gds
        return self.get_result()
    
    def run_innovus(self):
        pass


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   dse_facade.py
@Time    :   2024-08-29 10:54:34
@Author  :   SivanLaai
@Version :   1.0
@Contact :   lyhhap@163.com
@Desc    :   dse facade
'''

import sys
import os

current_dir = os.path.split(os.path.abspath(__file__))[0]
tool_dir = current_dir.rsplit('/', 2)[0]
print(tool_dir)
sys.path.append(tool_dir)
import time
import logging
import datetime
import os
import sys
import json
from application.DSE.arguements import Arguements
from application.DSE.config_management import ConfigManagement
from database.enum import EdaTool, FlowStep, DSEMethod
from application.DSE.optimization import WandbOptimization, NNIOptimization
from workspace.path import WorkspacePath
import wandb
import yaml


class DSEFacade:
    params = None
    def objective(self, trial):
        return self.start(trial=trial)

    def run_optuna(self):
        study = None
        sampler = optuna.samplers.TPESampler()
        study = optuna.create_study(storage="mysql+pymysql://username:password@ip:port/optuna", study_name=self.args.project_name, \
                                     directions=["minimize", "maximize", "maximize"], sampler=sampler, load_if_exists=True)
        if not study:
            return
        study.optimize(self.objective, n_trials=100, timeout=None)

        pruned_trials = study.get_trials(deepcopy=False, states=[TrialState.PRUNED])
        complete_trials = study.get_trials(deepcopy=False, states=[TrialState.COMPLETE])

        print("Study statistics: ")
        print("  Number of finished trials: ", len(study.trials))
        print("  Number of pruned trials: ", len(pruned_trials))
        print("  Number of complete trials: ", len(complete_trials))

        print("Best trial:")
        trials = study.best_trials

        for trial in trials:
            print("  Value: ", trial)

            print("  Params: ")
            for key, value in trial.params.items():
                print("    {}: {}".format(key, value))

    def run_nni(self, algorithm="TPE", direction="minimize", search_space=dict(), concurrency=1, max_trial_number=2000, flows=None):
        from nni.experiment import Experiment
        import random
        experiment = Experiment('local')
        port = 8088
        # print(vars(self.args))


        try:
            # trial_command = f'python optimization.py --tuner nni --step {self.step.value.lower()} --flows {flows}'
            arg_setting = ""
            for k,v in vars(self.args).items():
               if isinstance(v, bool):
                   if v:
                       arg_setting += f" --{k}"
               elif v is not None:
                   arg_setting += f" --{k} {v}"
            # print(arg_setting)
            trial_command = f'python optimization.py {arg_setting}'
            print(trial_command)
            experiment.config.trial_command = trial_command 
            experiment.config.trial_code_directory = os.path.dirname(__file__)
            experiment.config.search_space = search_space

            experiment.config.tuner.name = algorithm
            experiment.config.tuner.class_args['optimize_mode'] = direction

            experiment.config.max_trial_number = self.args.run_count
            experiment.config.trial_concurrency = self.args.sweep_worker_num
            experiment.run(port)
            # experiment.run(random.Random().randint(3000, 60036))
        except Exception as e:
            print(e)
            port = random.Random().randint(3000, 60036)
            experiment.run(port)
            # print(arg_setting)

    def start(self, optimize=DSEMethod.WANDB, eda_tool=EdaTool.IEDA, step=FlowStep.place):
        self.args = Arguements.parse(sys.argv[1:])
        config_manage = ConfigManagement(self.args, eda_tool)
        self.step = config_manage.getStep()
        # self.config_paths = config_manage.getConfigPathList()
        dir_workspace = config_manage.getWorkspacePath()

        self.params = config_manage.getParameters()

        os.environ["OMP_NUM_THREADS"] = "%d" % (self.params.num_threads)

        if optimize==DSEMethod.WANDB:
            self.setEnv(self.args)
            method = WandbOptimization(self.args, dir_workspace, self.params, self.step)
            method.runOptimization(tool=parameter)
        elif optimize==DSEMethod.NNI:
            method = NNIOptimization(self.args, dir_workspace, self.params, self.step)
            self._search_space = method._search_config
            print(self._search_space)
            self.run_nni(search_space=self._search_space, flows=self.args.flows)
        elif optimize==DSEMethod.OPTUNA:
            pass

    def setEnv(self, args):
        os.environ["WANDB_PROJECT"] = args.project_name
        os.environ["WANDB_MODE"] = args.wandb_mode



if __name__ == "__main__":
    auto = DSEFacade(None) 
    auto.start(optimize=DSEMethod.NNI)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : timing_opt.py
@Author : yell
@Desc : timing optimization framework
'''
from engine.base import EngineBase
from database.enum import EdaTool, FlowStep
from flow.flow_db import DbFlow
from tools.iEDA.module.timing_opt import IEDATimingOpt

class EngineTimingOpt(EngineBase):
    """timing optimization framework"""   
    
    #######################################################################################    
    # timing optimization for drv
    #######################################################################################
    def run_to_drv(self):
        if(self.eda_tool == EdaTool.IEDA):
            return self.run_to_drv_ieda()
            
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.run_to_drv_innovus()
        
        
    def run_to_drv_ieda(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                            step=FlowStep.cts)
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                        pre_flow=self.pre_step,
                        step=FlowStep.optDrv)
        self.build_path()
        
        # run iEDA
        run_flow = IEDATimingOpt(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_to_drv()
        
        return self.run_feature_ieda()
    
    def run_to_drv_innovus(self):
        pass
    
    #######################################################################################    
    # timing optimization for hold
    #######################################################################################  
    def run_to_hold(self):
        if(self.eda_tool == EdaTool.IEDA):
            return self.run_to_hold_ieda()
            
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.run_to_hold_innovus()
        
    def run_to_hold_ieda(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                            step=FlowStep.optDrv)
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                      pre_flow=self.pre_step,
                      step=FlowStep.optHold)
        self.build_path()
        
        # run iEDA
        run_flow = IEDATimingOpt(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_to_hold()
        
        return self.run_feature_ieda()
    
    def run_to_hold_innovus(self):
        pass
    
    #######################################################################################    
    # timing optimization for setup
    #######################################################################################  
    def run_to_setup(self):
        if(self.eda_tool == EdaTool.IEDA):
            return self.run_to_setup_ieda()
            
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.run_to_setup_innovus()
        
    def run_to_setup_ieda(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                              step=FlowStep.optHold)
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                      pre_flow=self.pre_step,
                      step=FlowStep.optSetup)
        self.build_path()
        
        # run iEDA
        run_flow = IEDATimingOpt(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_to_setup()
        
        return self.run_feature_ieda()
    
    def run_to_setup_innovus(self):
        pass
    


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : floorplan.py
@Author : yell
@Desc : floorplan framework
'''
import os

from engine.base import EngineBase
from database.enum import EdaTool, FlowStep
from flow.flow_db import DbFlow
from tools.iEDA.module.floorplan import IEDAFloorplan
from tools.innovus.module.floorplan import InnovusFloorplan

if os.environ.get('OpenROAD') == "on":
    from tools.OpenROAD.module.floorplan import OpenROADFloorplan

class EngineFloorplan(EngineBase):
    """floorplan framework"""
    def run(self):
        return super().run()
        
    
    def run_ieda(self):
        # build flow
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                      step=FlowStep.floorplan)
        self.build_path()
        
        # run iEDA
        run_flow = IEDAFloorplan(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_floorplan()
        
        return self.run_feature_ieda()
    
    def run_innovus(self):
        # build flow
        self.step = DbFlow(eda_tool=EdaTool.INNOVUS,
                      step=FlowStep.floorplan)
        self.build_path()
        
        # run
        run_flow = InnovusFloorplan(dir_workspace = self.dir_workspace,
                               dir_resource = self.dir_resource,
                              input_def = self.input_def, 
                              input_verilog = self.input_verilog,
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              step = self.step,
                              task = self.task)
        run_flow.run_floorplan()
        
        # update result
        return self.get_result()
    
    def run_OpenROAD(self):
        # build flow
        self.step = DbFlow(eda_tool=EdaTool.OPENROAD,
                      step=FlowStep.floorplan)
        self.build_path()
        
        # run
        run_flow = OpenROADFloorplan(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              input_verilog = self.input_verilog,
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              step = self.step,
                              task = self.task)
        run_flow.run_floorplan()
        
        # update result
        return self.get_result()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : lm_graph.py
@Author : yell
@Desc : feature for large model
'''

from database.enum import EdaTool, FeatureOption, FlowStep
from workspace.path import WorkspacePath, LargeModelFeatureType
from flow.flow_db import DbFlow
from database.large_model import *
from typing import List, Tuple
import numpy as np
import pandas as pd
import logging
from joblib import Parallel, delayed
from tqdm import tqdm
from numba import njit, prange
from utility.json_parser import JsonParser
import os


def _parse_patch(json_path: str) -> LmPatch:
    lm_patch = LmPatch()

    json_parser = JsonParser(json_path)
    if json_parser.read():
        json_patch = json_parser.get_json_data()

        # patch
        lm_patch.id = json_patch.get('id')
        lm_patch.patch_id_row = json_patch.get('patch_id_row')
        lm_patch.patch_id_col = json_patch.get('patch_id_col')
        lm_patch.llx = json_patch.get('llx')
        lm_patch.lly = json_patch.get('lly')
        lm_patch.urx = json_patch.get('urx')
        lm_patch.ury = json_patch.get('ury')
        lm_patch.row_min = json_patch.get('row_min')
        lm_patch.row_max = json_patch.get('row_max')
        lm_patch.col_min = json_patch.get('col_min')
        lm_patch.col_max = json_patch.get('col_max')
        lm_patch.area = json_patch.get('area')
        lm_patch.cell_density = json_patch.get('cell_density')
        lm_patch.pin_density = json_patch.get('pin_density')
        lm_patch.net_density = json_patch.get('net_density')
        lm_patch.macro_margin = json_patch.get('macro_margin')
        lm_patch.RUDY_congestion = json_patch.get('RUDY_congestion')
        lm_patch.EGR_congestion = json_patch.get('EGR_congestion')
        lm_patch.timing_map = json_patch.get('timing_map')
        lm_patch.power_map = json_patch.get('power_map')
        lm_patch.ir_drop_map = json_patch.get('ir_drop_map')   

        # patch layer
        json_patch_layers = json_patch.get('patch_layer', [])
        for json_patch_layer in json_patch_layers:
            patch_layer = LmPatchLayer()
            patch_layer.id = json_patch_layer.get('id')
            patch_layer.net_num = json_patch_layer.get('net_num')
            feature = json_patch_layer.get('feature', {})
            patch_layer.wire_width = feature.get('wire_width')
            patch_layer.wire_len = feature.get('wire_len')
            patch_layer.wire_density = feature.get('wire_density')
            patch_layer.congestion = feature.get('congestion')

            json_nets = json_patch_layer.get('nets', [])
            for json_net in json_nets:
                # net
                lm_net = LmNet()
                lm_net.id = json_net.get('id')
                lm_net.name = json_net.get('name')

                # wires
                lm_net.wire_num = json_net.get('wire_num')
                json_wires = json_net.get('wires', [])
                for json_wire in json_wires:
                    lm_wire = LmWire()
                    lm_wire.id = json_wire.get('id')

                    # wire feature
                    wire_feature = LmWireFeature()
                    feature = json_wire.get('feature', {})
                    wire_feature.wire_len = feature.get('wire_len')

                    lm_wire.feature = wire_feature

                    # path
                    lm_wire.path_num = json_wire.get('path_num')
                    json_paths = json_wire.get('paths', [])
                    for json_path in json_paths:
                        wire_path = LmPath()

                        lm_path_node1 = LmNode()
                        lm_path_node1.id = json_path.get('id1')
                        lm_path_node1.x = json_path.get('x1')
                        lm_path_node1.y = json_path.get('y1')
                        lm_path_node1.row = json_path.get('r1')
                        lm_path_node1.col = json_path.get('c1')
                        lm_path_node1.layer = json_path.get('l1')
                        lm_path_node1.pin_id = json_path.get('p1')
                        wire_path.node1 = lm_path_node1

                        lm_path_node2 = LmNode()
                        lm_path_node2.id = json_path.get('id2')
                        lm_path_node2.x = json_path.get('x2')
                        lm_path_node2.y = json_path.get('y2')
                        lm_path_node2.row = json_path.get('r2')
                        lm_path_node2.col = json_path.get('c2')
                        lm_path_node2.layer = json_path.get('l2')
                        lm_path_node2.pin_id = json_path.get('p2')
                        wire_path.node2 = lm_path_node2

                        lm_wire.paths.append(wire_path)

                    lm_net.wires.append(lm_wire)

                patch_layer.nets.append(lm_net)

            lm_patch.patch_layer.append(patch_layer)

    return lm_patch


class LmPatchs():
    """feature statis"""

    def __init__(self, dir_workspace: str, eda_tool: EdaTool):
        self.dir_workspace = dir_workspace
        self.eda_tool = eda_tool
        self.workspace = WorkspacePath(dir_workspace)

    def get_patches(self) -> List[LmPatch]:
        lm_patches = []

        flow = DbFlow(eda_tool=self.eda_tool, step=FlowStep.route)
        feature_dir = self.workspace.get_feature_lm(flow=flow,
                                                    feature_option=FeatureOption.large_model,
                                                    lm_feature_type=LargeModelFeatureType.lm_patches)
        logging.info("Start to read patches...")
        for root, _, files in os.walk(feature_dir):
            # lm_patches = Parallel(backend="loky")(delayed(_parse_patch)(os.path.join(root, file))
            #                                       for file in tqdm(files, desc="Reading patches"))
            lm_patches = [_parse_patch(os.path.join(root, file))
                          for file in tqdm(files, desc="Reading patches")]

        return lm_patches


@dataclass
class PathData:
    channel_id: int
    node1_x: float
    node1_y: float
    node2_x: float
    node2_y: float


@dataclass
class PatchData:
    patch_id_row: int
    patch_id_col: int
    lly: int
    ury: int
    llx: int
    urx: int
    paths: List[PathData]


@njit(parallel=True)
def _resize_3D_matrix_numba(matrix: np.ndarray, size: Tuple[int, int]) -> np.ndarray:
    """
    Numba-accelerated function to resize a 3D matrix by resizing each 2D channel.

    Args:
        matrix (np.ndarray): The input 3D matrix.
        size (Tuple[int, int]): The output size.
    """
    size_x, size_y = size
    channels, orig_x, orig_y = matrix.shape
    result = np.zeros((channels, size_x, size_y), dtype=np.float64)

    # Precompute block sizes and start indices for both dimensions
    x_block_sizes = np.full(size_x, orig_x // size_x, dtype=np.int32)
    for i in range(orig_x % size_x):
        x_block_sizes[i] += 1

    y_block_sizes = np.full(size_y, orig_y // size_y, dtype=np.int32)
    for j in range(orig_y % size_y):
        y_block_sizes[j] += 1

    x_start_indices = np.zeros(size_x, dtype=np.int32)
    y_start_indices = np.zeros(size_y, dtype=np.int32)

    for i in range(1, size_x):
        x_start_indices[i] = x_start_indices[i - 1] + x_block_sizes[i - 1]

    for j in range(1, size_y):
        y_start_indices[j] = y_start_indices[j - 1] + y_block_sizes[j - 1]

    # Iterate over each channel
    for c in prange(channels):
        # Iterate over each block in x and y dimensions
        for i in range(size_x):
            for j in range(size_y):
                x_start = x_start_indices[i]
                x_end = x_start + x_block_sizes[i]
                y_start = y_start_indices[j]
                y_end = y_start + y_block_sizes[j]

                sum_val = 0.0
                count = 0
                for xi in range(x_start, x_end):
                    for yi in range(y_start, y_end):
                        sum_val += matrix[c, xi, yi]
                        count += 1

                if count > 0:
                    result[c, i, j] = sum_val / count
                else:
                    result[c, i, j] = 0.0

        # Normalize each channel by its maximum value
        max_val = 0.0
        for i in range(size_x):
            for j in range(size_y):
                if result[c, i, j] > max_val:
                    max_val = result[c, i, j]

        if max_val > 0.0:
            for i in range(size_x):
                for j in range(size_y):
                    result[c, i, j] /= max_val

    return result


def _process_patch(n_channels: int, roughness: int, patch_data: PatchData) -> Tuple[int, int, np.ndarray]:
    # Initialize the matrix for this patch
    patch_width = (patch_data.urx - patch_data.llx) // roughness
    patch_height = (patch_data.ury - patch_data.lly) // roughness

    matrix = np.zeros((n_channels, patch_height, patch_width))

    # Iterate over each channel
    for path in patch_data.paths:
        x1 = (path.node1_x - patch_data.llx) // roughness
        x2 = (path.node2_x - patch_data.llx) // roughness
        y1 = (path.node1_y - patch_data.lly) // roughness
        y2 = (path.node2_y - patch_data.lly) // roughness
        x1, x2 = min(x1, x2), max(x1, x2)
        y1, y2 = min(y1, y2), max(y1, y2)
        # Update the matrix
        matrix[path.channel_id, y1:y2 + 1, x1:x2 + 1] += 1

    return (patch_data.patch_id_row, patch_data.patch_id_col, matrix)


def _process_resized_patch(n_channels: int, roughness: int, patch_data: PatchData, patch_size: Tuple[int, int]) -> Tuple[int, int, np.ndarray]:
    row, col, matrix = _process_patch(n_channels, roughness, patch_data)

    return (row, col, _resize_3D_matrix_numba(matrix, patch_size))


class LmPatchOperator():
    def __init__(self, roughness: int = 1, ignore_empty: bool = False, patch_size: Tuple[int, int] = None):
        # (row, col) -> (#layer/channel, height, width)
        self._matrices: List[List[np.ndarray]] = []
        self._channels: List[int] = []
        self._roughness: int = roughness
        self._ignore_empty: bool = ignore_empty
        self._patch_size: Tuple[int, int] = patch_size

    def get_channels(self) -> List[int]:
        return self._channels

    def to_matrix(self, patches: List[LmPatch]) -> List[List[np.ndarray]]:
        """
        Convert a list of LmPatch objects into a #layer-channel image representation.
        Each channel represents a wire_feature (e.g., wire_width, wire_len, etc.).

        Args:
            patches (List[LmPatch]): The input list of patch objects.

        Returns:
            #layer-channel images map (row,col) -> (#layer, height, width).
            List[List[np.ndarray]]: The output list of matrices.
        """

        logging.info("Start to convert patch to image...")
        if len(patches) == 0:
            logging.warning("No patches found, aborting...")
            return None
        # Pre-process
        self._find_feasible_channels(patches)

        # Determine matrix dimensions
        n_row = max(patch.patch_id_row for patch in patches) + 1
        n_col = max(patch.patch_id_col for patch in patches) + 1

        # Initialize matrices
        self._matrices = [
            [None for _ in range(n_col)] for _ in range(n_row)]
        patches_data = [self._extract_patch_data(
            patch) for patch in patches]

        if self._patch_size is None:
            results = Parallel(backend="threading")(
                delayed(_process_patch)(len(self._channels), self._roughness, patch_data) for patch_data in tqdm(patches_data, desc="Processing patches")
            )
        else:
            results = Parallel(backend="threading")(
                delayed(_process_resized_patch)(len(self._channels), self._roughness, patch_data, self._patch_size) for patch_data in tqdm(patches_data, desc="Processing patches")
            )
        # Assign the processed matrices to their respective positions
        for row, col, matrix in results:
            self._matrices[row][col] = matrix

        # Count the number of processed patches
        processed_patches = len(results)

        logging.info(
            f"> #Patches: {processed_patches}, #Rows: {n_row}, #Cols: {n_col}, (Resolution: {self._roughness})")

        return self._matrices

    def distribution_analysis(self, csv_path: str = None) -> pd.DataFrame:
        """
        Analyze the distribution of routing wire

        Args:
            csv_path (str): The output CSV file path.

        Returns:
            pd.DataFrame: The output DataFrame.
        """
        logging.info("Start to analyze the distribution of routing wire...")

        # Initialize lists to store results
        patch_id_rows, patch_id_cols, layer_ids, sums, densities = [], [], [], [], []

        # Precompute sizes to avoid repeated calculations in loops
        n_row, n_col = len(self._matrices), len(self._matrices[0])
        for row_id in range(n_row):
            for col_id in range(n_col):
                matrix = self._matrices[row_id][col_id]
                if matrix is None:
                    continue

                # Vectorized calculation of sums and densities
                layer_sums = np.sum(matrix, axis=(1, 2))
                layer_size = matrix.shape[1] * matrix.shape[2]
                layer_densities = layer_sums / layer_size

                # Use numpy to quickly extend lists
                num_layers = len(layer_sums)
                patch_id_rows.extend([row_id] * num_layers)
                patch_id_cols.extend([col_id] * num_layers)
                layer_ids.extend(range(num_layers))
                sums.extend(layer_sums)
                densities.extend(layer_densities)

        # Convert lists to a DataFrame
        df = pd.DataFrame({
            'patch_id_row': np.array(patch_id_rows, dtype=np.int32),
            'patch_id_col': np.array(patch_id_cols, dtype=np.int32),
            'layer_id': np.array(layer_ids, dtype=np.int32),
            'sum': np.array(sums, dtype=np.int32),
            'density': np.array(densities, dtype=np.float32),
        })

        # Calculate and log zero statistics
        zero_mask = (df['sum'] == 0)
        zero_count = zero_mask.sum()
        total_count = len(df)
        zero_ratio = zero_count / total_count * 100
        logging.info(
            f"> #Empty wire matrix: {zero_count} / {total_count}, ratio: {zero_ratio:.2f}%"
        )

        # Group by 'layer_id' and calculate stats
        layer_group = df.groupby('layer_id')['sum']
        count_by_layer = layer_group.size()
        zero_by_layer = layer_group.apply(lambda x: (x == 0).sum())
        ratio_by_layer = zero_by_layer / count_by_layer * 100

        # Log layer-specific statistics
        for layer_id, total_num in count_by_layer.items():
            z_count = zero_by_layer[layer_id]
            z_ratio = ratio_by_layer[layer_id]
            logging.info(
                f"> Layer {layer_id}: #Empty wire matrix: {z_count} / {total_num}, ratio: {z_ratio:.2f}%"
            )

        # Save to CSV if path is provided
        if csv_path:
            df.to_csv(csv_path, index=False)
            logging.info(f"Distribution result has been saved to {csv_path}.")

        return df

    def _find_feasible_channels(self, patches: List[LmPatch]):
        """
        Find feasible channels from a list of LmPatch objects.

        Args:
            patches (List[LmPatch]): The input list of patch objects.
        """
        logging.info("Start to find feasible channels...")
        # Find feasible channels
        original_channels = [
            layer.id for layer in patches[0].patch_layer]
        logging.info(f"> #Original Total channels: {len(original_channels)}")

        if not self._ignore_empty:
            self._channels = original_channels
        else:
            for channel in original_channels:
                is_empty = True
                for patch in patches:
                    for layer in patch.patch_layer:
                        if layer.id != channel:
                            continue

                        for net in layer.nets:
                            net: LmNet
                            if (len(net.wires) > 0):
                                is_empty = False
                                break
                if not is_empty:
                    self._channels.append(channel)
                else:
                    logging.warning(
                        f"Channel {channel} is empty and will be ignored")

        logging.info(f"> #Feasible Total channels: {len(self._channels)}")

    def _extract_patch_data(self, patch: LmPatch) -> PatchData:
        """
        Extract the patch data from an LmPatch object.

        Args:
            patch (LmPatch): The input patch object.

        Returns:
            PatchData: The output patch data.
        """
        paths = []
        for layer_id in self._channels:
            layer = patch.patch_layer[layer_id]
            for net in layer.nets:
                for wire in net.wires:
                    for path in wire.paths:
                        node1, node2 = path.node1, path.node2
                        if node1 is None or node2 is None:
                            continue
                        path_data = PathData(
                            channel_id=layer_id,
                            node1_x=node1.x,
                            node1_y=node1.y,
                            node2_x=node2.x,
                            node2_y=node2.y
                        )
                        paths.append(path_data)

        return PatchData(
            patch_id_row=patch.patch_id_row,
            patch_id_col=patch.patch_id_col,
            lly=patch.lly,
            ury=patch.ury,
            llx=patch.llx,
            urx=patch.urx,
            paths=paths
        )



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
'''
@File : design.py
@Author : yell
@Desc : design database
'''

import sys
import os

from dataclasses import dataclass
from dataclasses import field

from numpy import uint64

from database.enum import CellType, OrientType, PlaceStatus, NetType

@dataclass     
class Design(object):
     """basic feature package"""
     instances : list = field(default_factory=list)
     nets : list = field(default_factory=list)
     
    
@dataclass     
class Instance(object):
    """instance structure"""
    name : str = ""
    master : str = ""
    type : CellType = CellType.none
    llx : int = 0
    lly : int = 0
    urx : int = 0
    ury : int = 0
    orient : OrientType = OrientType.none
    status : PlaceStatus = PlaceStatus.none
    pins : list = field(default_factory=list)
    
@dataclass     
class Pin(object):
    """pin structure"""
    name : str = ""
    instance : str = ""
    net : str = ""
    center_x : int = 0
    center_y : int = 0
    
@dataclass     
class Net(object):
    """net structure"""
    name : str = ""
    type : NetType = NetType.none
    pin_number : uint64 = None
    wire_len : uint64 = None
    segment_number : uint64 = None
    via_number : uint64 = None
    pins : list = field(default_factory=list)
    
class NetPin(object):
    """pin in net"""
    pin_name : str = ""
    instance_name : str = ""





import argparse

class Arguements:
    @classmethod
    def parse(cls, sys_args):
        parser = argparse.ArgumentParser(
            description='AiEDA argument management', formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument("--team_name", type=str, default="laai")
        parser.add_argument("--project_name", type=str, default="gcd")
        parser.add_argument("--experiment_name", type=str, default=None)
        parser.add_argument("--scenario_name", type=str, default="test_sweep")
        parser.add_argument("--wandb_log_path", type=str, default="./wandb_results/")
        parser.add_argument("--wandb_mode",type=str, choices=["offline", "online"],default="offline", help="[WANDB] wandb optimization mode")
        parser.add_argument("--seed",type=int,default=0)
        parser.add_argument("--sweep_worker_num",type=int,default=1)
        parser.add_argument("--num_threads",type=int,default=1)
        parser.add_argument("--config_path",type=str,default="/data/laixinhua/AiEDA/application/benchmark/28nm/gcd/config/iEDA_config/cts_default_config.json")
        parser.add_argument("--flows",type=str,default=None)
        parser.add_argument("--step",type=str,default="full_flow", help='''
                                NoStep = ""
                                initFlow = "initFlow"
                                initDB = "initDB"
                                edi = "edi"
                                floorplan = "floorplan"
                                pdn = "PDN"
                                prePlace = "prePlace"
                                place = "place"
                                globalPlace = "gp"
                                detailPlace = "dp"
                                cts = "CTS"
                                route = "route"
                                fixFanout = "fixFanout"
                                optDrv = "optDrv"
                                optHold = "optHold"
                                optSetup = "optSetup"
                                legalization = "legalization"
                                full_flow = 'full_flow'
                                custom_flow = 'custom_flow'
                            ''')
        parser.add_argument("--result_dir",type=str,default="./result")
        parser.add_argument("--gpu",type=int,default=0)
        parser.add_argument("--disable_gpu",type=int,default=None)
        parser.add_argument("--gpu_analyze",type=int,default=0)
        parser.add_argument("--sweep_id",type=str,default=None)
        parser.add_argument("--run_count",type=int,default=10)
        parser.add_argument("--plot_flag",type=int,default=0)
        parser.add_argument("--tuner",type=str,default="wandb", choices=["wandb", "nni", "optuna"], help="setting tuner method")
        parser.add_argument("--multobj_flag",type=int,default=0, choices=[1, 0], help="multiple objective optimization")
        parser.add_argument("--store_ref",type=int,default=0, choices=[1, 0], help="store current metric to ref")
        parser.add_argument("--root", type=str, default="/data/laixinhua/AiEDA/application/benchmark", help="workspace root diretory name")
        parser.add_argument("--tech", type=str, default="28nm", help="technology design")
        parser.add_argument("--benchmark_flag", type=bool, default=False, help="weather to generate benchmark")
        parser.add_argument("--benchmark_output", type=str, default="benchmark_output", help="benchmark output path")
        all_args = parser.parse_known_args(sys_args)[0]
        return all_args


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : rcx.py
@Author : yexinyu
@Desc : RCX framenwork
'''

import sys 
sys.path.insert(0, r'/home/yexinyu/AiEDA')
sys.path.insert(0, r'/home/yexinyu/AiEDA/database/enum.py')
import os
from engine.base import EngineBase
from database.enum import EdaTool, FlowStep
# if os.environ.get('StarRC') == "on":
from tools.StarRC.module.rcx import StarRCRcx 

class EngineRCX(EngineBase):
    """routing framenwork"""   
    def __init__(self, dir_workspace, dir_resource = "", eda_tool = EdaTool.STARRC, input_def = "", block = "" , output_spef=None
                #  step = FlowStep.rcx
                 ):
        super().__init__(dir_workspace, dir_resource, eda_tool, input_def, block, output_spef)

    def run(self):
        if(self.step == FlowStep.rcx):
            return self.run_rcx()

    #######################################################################################    
    # rcx
    #######################################################################################
    def run_rcx(self, input_def, block): 
        if(self.eda_tool == EdaTool.STARRC):
            return self.run_starrc_rcx(input_def, block)
        # elif(self.eda_tool == EdaTool.INNOVUS):
        #     return self.run_innovus_rcx()
        else:
            print("Error, don't support this tool : " + self.eda_tool.value)
            exit(0)
        

    def run_starrc_rcx(self,input_def, block):
            
        # self.build_path()
        
        # run StarRC
        run_flow = StarRCRcx(dir_workspace = self.dir_workspace,
                            dir_resource = self.dir_resource,
                            input_def = input_def, 
                            input_verilog = self.input_verilog,
                            output_spef = self.output_spef, 
                            output_verilog = self.output_verilog,
                            pre_step = self.pre_step,
                            step = self.step,
                            task = self.task)
    
   
        run_flow.run_rcx(input_def, block)
        return self.get_result()
    



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
'''
@File : layout.py
@Author : yell
@Desc : layout database
'''

import sys
import os

from dataclasses import dataclass
from dataclasses import field

from database.enum import TrackDirection, LayerType

@dataclass     
class Layout(object):
     """basic feature package"""
     top_name : str = ""
     dbu : int = 0
     die = None
     core = None
     rows = None
     tracks : list = field(default_factory=list)
    #  layers : list = field(default_factory=list)
     routing_layers : list = field(default_factory=list)
    #  cut_layers : list = field(default_factory=list)

@dataclass     
class Die(object):
    """die structure"""
    llx : int = 0
    lly : int = 0
    urx : int = 0
    ury : int = 0
    
@dataclass     
class Core(object):
    """core structure"""
    llx : int = 0
    lly : int = 0
    urx : int = 0
    ury : int = 0
    
@dataclass     
class Rows(object):
    """row structure"""
    num_rows : int = 0
    row_width : int = 0
    row_height : int = 0
    
class Track(object):
    """track structure"""
    layer : str = ""
    prefer_dir : TrackDirection = TrackDirection.none
    num : int = 0
    start : int = 0
    step : int = 0
    
class Layer(object):
    """layer structure"""
    name : str = ""
    type : LayerType = LayerType.none
    min_width : int = 0
    max_width : int = 0
    width : int = 0
    area : int = 0


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : flow_manager.py
@Author : yell
@Desc : Manage flow
'''
from engine.base import EngineBase

class EngineDatabase(EngineBase):
    """Manage flow"""
    def __init__(self, dir_workspace : str):
        super().__init__(dir_workspace)
        
    
    def run(self):
        pass


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : full_flow.py
@Author : yell
@Desc : full flow engine api
'''
from engine.base import EngineBase
from flow.flow_db import DbFlow
from database.enum import FlowStep, EdaTool
from tools.innovus.module.full_flow import InnovusFullFlow
from feature.io import FeatureIO

class EngineFullFlow(EngineBase):
    """fullflow engine api"""
    def run(self):
        return super().run()
    
    def run_innovus(self):
        # build flow
        self.step = DbFlow(eda_tool=EdaTool.INNOVUS,
                          step=FlowStep.full_flow)
            
        self.build_path()
        
        # run innovus
        run_flow = InnovusFullFlow(dir_workspace = self.dir_workspace,
                               dir_resource = self.dir_resource,
                              input_def = self.input_def, 
                              input_verilog = self.input_verilog,
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              step = self.step,
                              task = self.task)
        run_flow.run_full_flow()
        
        # update result
        # self.result.output_features['summary_cts'] = feature_path
        
        return self.get_result()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : fpt.py
@Time : 2022/05/19 11:42:44
@Author : simin tao
@Version : 1.0
@Contact : taosm@pcl.ac.cn
@Desc : The top of feature parser tools.
'''



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : drc.py
@Author : yell
@Desc : drc framework
'''
from engine.base import EngineBase
from flow.flow_db import DbFlow
from database.enum import FlowStep, EdaTool
from tools.iEDA.module.drc import IEDADrc
from tools.iEDA.feature.feature import IEDAFeature
from tools.innovus.module.drc import InnovusDRC

class EngineDRC(EngineBase):
    """drc framework"""
    def run(self):
        return super().run()
        
    def run_ieda(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                            step=FlowStep.route)
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                      pre_flow=self.pre_step,
                      step=FlowStep.drc)
        self.build_path()
        
        # run iEDA drc
        run_flow = IEDADrc(dir_workspace = self.dir_workspace,
                              input_def = self.input_def)
        run_flow.run_drc()
        
        return None
    
    def run_innovus(self):
        # build flow
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.INNOVUS,
                              step=FlowStep.route)
        if(self.step == None):
            self.step = DbFlow(eda_tool=EdaTool.INNOVUS,
                          pre_flow=self.pre_step,
                          step=FlowStep.drc)
            
        self.build_path()
        
        # run innovus
        run_flow = InnovusDRC(dir_workspace = self.dir_workspace,
                               dir_resource = self.dir_resource,
                              input_def = self.input_def, 
                              input_verilog = self.input_verilog,
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              pre_step = self.pre_step,
                              step = self.step,
                              task = self.task)
        run_flow.run_drc()
        
        return self.get_result()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : jsonl_parser.py
@Author : Yihang Qiu
@Desc : jsonl parser 
'''

import json
import gzip
import os
import logging


class JsonlParser:
    """basic jsonl parser"""

    def __init__(self, jsonl_path: str):
        self.jsonl_path = jsonl_path
        self.jsonl_data = []

    def set_jsonl_data(self, value: list):
        self.jsonl_data = value

    def get_jsonl_data(self):
        ''' access to jsonl data '''
        return self.jsonl_data

    def get_value(self, node_json, key):
        if key in node_json:
            return node_json[key]
        else:
            return None

    def read(self):
        ''' Jsonl reader '''
        if not os.path.exists(self.jsonl_path):
            logging.info("error : jsonl file not exist. path = %s",
                         self.jsonl_path)
            return False
        else:
            logging.info("info : start parsing file : %s", self.jsonl_path)

        try:
            # compressed
            if self.jsonl_path.endswith(".gz"):
                with gzip.open(self.jsonl_path, 'rt', encoding='utf-8') as f:
                    self.jsonl_data = [json.loads(line.strip()) for line in f if line.strip()]
            else:
                with open(self.jsonl_path, 'r', encoding='utf-8') as f_reader:
                    self.jsonl_data = [json.loads(line.strip()) for line in f_reader if line.strip()]
            
            logging.info("info : parsing file success")
            return True
        except json.JSONDecodeError:
            logging.info(
                "error : jsonl file format error. path = %s", self.jsonl_path)
            return False

    def read_create(self):
        if not os.path.exists(self.jsonl_path):
            # create file
            self.jsonl_data = []
            self.write()

        return self.read()

    def write(self, list_value: list = None):
        if list_value is not None:
            self.jsonl_data = list_value

        ''' Jsonl writer '''
        if self.jsonl_path.endswith(".gz"):
            with gzip.open(self.jsonl_path, 'wt', encoding='utf-8') as f:
                for item in self.jsonl_data:
                    f.write(json.dumps(item) + '\n')
        else:
            with open(self.jsonl_path, 'w', encoding='utf-8') as f_writer:
                for item in self.jsonl_data:
                    f_writer.write(json.dumps(item) + '\n')

        logging.info("info : write file success")
        return True

    def get_value(self, dict_node, key):
        if isinstance(key, str) and key in dict_node:
            return dict_node[key]

        if isinstance(key, list):
            dict_word = dict_node
            for word in key:
                if word in dict_word:
                    dict_word = dict_word[word]
                else:
                    return None

            return dict_word

        return None



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : base.py
@Author : yell
@Desc : an EDA engine framework 
'''
import sys 
# sys.path.insert(0, r'/home/yexinyu/AiEDA/engine/base.py')
# sys.path.insert(0, r'/home/yexinyu/AiEDA')

from database.enum import EdaTool, FlowStep, TaskOption, FeatureOption
from settings.setting import Settings
from workspace.path import WorkspacePath
from flow.flow_db import DbFlow
from feature.io import FeatureIO
from typing import List


class EngineResult(object):
    """data structure for output"""
    intput_def : str = ""
    input_verilog : str = ""
    output_def : str = ""
    output_verilog : str = ""
    output_gds = ""
    output_features = {}
    output_logs = {}
    output_reports = {}
    
    def print_data(self):
        print(self.intput_def)
        print(self.input_verilog)
        print(self.output_def)
        print(self.output_verilog)
        print(self.output_gds)
        print(self.output_features)
        print(self.output_logs)
        print(self.output_reports)

class EngineBase():
    """engine base"""
    
    def __init__(self, dir_workspace : str, dir_resource : str = "",
                 eda_tool : EdaTool = EdaTool.IEDA, task : TaskOption = TaskOption.RUN_EDA, output_spef : str = None,
                 input_def : str = None, input_verilog : str = None, 
                 output_def : str = None, output_verilog : str = None,
                 pre_step : DbFlow = None, step : DbFlow = None, first_arg : str = "starrc", input_params : List[str] = []):
        self.dir_workspace = dir_workspace
        self.dir_resource = dir_resource
        self.eda_tool = eda_tool
        self.task = task
        self.input_def = input_def
        self.input_verilog = input_verilog
        self.output_def = output_def
        self.output_verilog = output_verilog
        self.pre_step = pre_step
        self.step : DbFlow = step
        self.first_arg = first_arg
        self.input_params = input_params
        self.output_spef = output_spef
        
        if(self.step != None):
            self.step.pre_flow = pre_step

        self.workspace = WorkspacePath(dir_workspace)
        self.result = EngineResult()
        
    def build_path(self):   
        """rebuild path by flow step"""
        design_name = self.workspace.json_workspace.design
        
        # reset input def
        if(self.input_def == None):
            self.input_def =  self.workspace.get_input_def(self.step)
        
        # reset input verilog        
        if(self.input_verilog == None):
            self.input_verilog =  self.workspace.get_input_verilog(self.step)
                
        # reset output def
        if(self.output_def == None):
            self.output_def = self.workspace.get_output_def(self.step)
        
        # reset output verilog
        if(self.output_verilog == None):
            self.output_verilog = self.workspace.get_output_verilog(self.step)
        
    def run(self):
        #build flow       
        if(self.eda_tool == EdaTool.IEDA):
            return self.run_ieda()
            
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.run_innovus()
        
        if(self.eda_tool == EdaTool.OPENROAD):
            return self.run_OpenROAD()
        
        if(self.eda_tool == EdaTool.DREAMPLACE):
            return self.run_dreamplace()
        
        if(self.eda_tool == EdaTool.STARRC):
            return self.run_rcx()
        
    def run_feature_ieda(self, reload : bool = False):     
        # save feature        
        feature_io = FeatureIO(dir_workspace = self.dir_workspace,
                               eda_tool = EdaTool.IEDA,
                               feature_option = FeatureOption.summary,
                               flow = self.step)
        
        output_features = feature_io.generate_db_ieda(reload)
        self.result.output_features = output_features
        
        return self.get_result()
        
    def get_result(self):
        self.result.intput_def=self.input_def,
        self.result.input_verilog=self.input_verilog,
        self.result.output_def=self.output_def,
        self.result.output_verilog=self.output_verilog
        
        self.result.print_data()

        return self.result
            
    
    
    


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : net_opt.py
@Author : yell
@Desc : net optimization framework
'''
from engine.base import EngineBase
from database.enum import EdaTool, FlowStep
from flow.flow_db import DbFlow
from tools.iEDA.module.net_opt import IEDANetOpt

class EngineNetOpt(EngineBase):
    """net optimization framework"""
    def run(self):
        return super().run()
        
    def run_ieda(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                              step=FlowStep.floorplan,
                              is_first=True)
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                        pre_flow=self.pre_step,
                        step=FlowStep.fixFanout)
        self.build_path()
        
        # run iEDA
        run_flow = IEDANetOpt(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_fix_fanout()
        
        return self.run_feature_ieda()
    
    def run_innovus(self):
        pass


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : base.py
@Author : yell
@Desc : feature io 
'''
from feature.base import FeatureBase
from tools.iEDA.utility.base import IEDABase
from database.enum import EdaTool, FeatureOption, FlowStep
from workspace.path import WorkspacePath
from multiprocessing import Process
from flow.flow_db import DbFlow
from utility.json_parser import JsonParser
import json

from tools.iEDA.feature.feature import IEDAFeature
from tools.innovus.feature.feature import InnovusFeature

class FeatureIO(FeatureBase):
    """feature base"""                
    def generate_db_ieda(self, reload : bool = False):        
        # if reload = True, read def again by process
        if(reload == True):
            p = Process(target=self.generate_db_ieda_process, args=(self.workspace, 
                                                                    self.flow,
                                                                    reload,
                                                                    self.input_def))
            p.start()
            p.join()
            
            return None
        else:
            ieda_feature = IEDAFeature(self.dir_workspace)
            # get feature
            path_summary = ieda_feature.feature_summary(flow = self.flow)
            if(reload == False):
                path_tool = ieda_feature.feature_tool(flow = self.flow)
            
            # design_eval_path = ieda_feature.feature_eval_summary(flow = self.flow)
            # net_eval_path = ieda_feature.feature_net_eval(flow = self.flow)
            
            output_features = {}
            output_features['summary'] = path_summary
            if(reload == False):
                output_features['tools'] = path_tool
            
            # output_features['design_eval_path'] = design_eval_path
            # output_features['net_eval_path'] = net_eval_path
            
            return output_features
    
    def generate_db_ieda_process(self, workspace_path : WorkspacePath, 
                                 flow : DbFlow,
                                 reload : bool = False,
                                 input_def : str = ""):        
        ieda_feature = IEDAFeature(workspace_path.workspace)
        # if reload = True, read def again
        if(reload == True):
            if(input_def == None):
                input_def = workspace_path.get_output_def(flow)
                # input_def = workspace_path.get_output_def(flow, False)
            ieda_feature.read_def(input_def)
            
        # get feature
        path_summary = ieda_feature.feature_summary(flow = flow)
        
        if (flow.step == DbFlow.step.place):
            design_eval_path, net_eval_path = ieda_feature.feature_pl_eval_union(flow = flow)
            print("eval_path = ", design_eval_path)
            print("net_eval_path = ", net_eval_path)    
        elif (flow.step == DbFlow.step.cts):
            design_eval_path, net_eval_path = ieda_feature.feature_cts_eval_union(flow = flow)
            print("eval_path = ", design_eval_path)
            print("net_eval_path = ", net_eval_path)    

        print("summary_path = ", path_summary)

        if(reload == False):
            path_tool = ieda_feature.feature_tool(flow = flow)
            print("path_tool = ", path_tool)
           
    def get_raw_json(self, feature_option : FeatureOption = FeatureOption.NoFeature):
        try:
            if(feature_option != FeatureOption.NoFeature):
                self.feature_option = feature_option

            reader = IEDABase(self.dir_workspace)      
            raw_json_path = reader.workspace.get_feature_jsonl(self.flow, feature_option=self.feature_option)
            data = dict()
            with open(raw_json_path, "r") as f:
                lines = f.readlines()
                for line in lines:
                    line_data = json.loads(line.strip())
                    data.update(line_data)
        except Exception as e:
            pass
        finally:
            return data
    
    def gernerate_db_ieda_by_innovus(self):
        # place
        feature_innovus = InnovusFeature(self.dir_workspace)
        feature_innovus.feature_timing(flow=self.flow, feature_option=FeatureOption.baseline_sta)
        feature_innovus.feature_power(flow=self.flow, feature_option=FeatureOption.baseline_power)
    
    def get_db_ieda(self):
        if(self.flow == None):
            print("Error : flow must be set to get db.")
            return
        
        if(self.feature_option == FeatureOption.summary):
            reader = IEDAFeature(self.dir_workspace)      
            summary = reader.feature_db_summary(self.flow)
            
            return summary
        
        if(self.feature_option == FeatureOption.tools):
            reader = IEDAFeature(self.dir_workspace)      
            tool_feature = reader.feature_db_tools(self.flow)
            
            return tool_feature

        if(self.feature_option == FeatureOption.eval):
            reader = IEDAFeature(self.dir_workspace)      
            eval_feature = reader.feature_db_eval(self.flow)
            
            return eval_feature
        
        if(self.feature_option == FeatureOption.baseline_sta):
            reader = InnovusFeature(self.dir_workspace)     
            sta_feature = reader.feature_db_timing(flow=self.flow, feature_option=FeatureOption.baseline_sta)
            
            return sta_feature
        
        if(self.feature_option == FeatureOption.baseline_power):
            reader = InnovusFeature(self.dir_workspace)     
            power_feature = reader.feature_db_power(flow=self.flow, feature_option=FeatureOption.baseline_power)
            
            return power_feature
        
    def generate_db_innovus(self):
        feature = InnovusFeature(self.dir_workspace)
        
        # feature tools by innovus feature
        feature.feature_tools()
        
        # feature summary by iEDA
        self.generate_db_ieda(reload=True)
    
    def get_db_innovus(self):
        if(self.flow == None):
            print("Error : flow must be set to get db.")
            return
        
        if(self.feature_option == FeatureOption.summary):
            reader = InnovusFeature(self.dir_workspace)      
            summary = reader.feature_db_summary(self.flow)
            
            return summary
        
        if(self.feature_option == FeatureOption.tools):
            reader = InnovusFeature(self.dir_workspace)      
            tool_feature = reader.feature_db_tools(self.flow)
            
            return tool_feature
        
    def gernerate_route_data(self, json_path : str, reload : bool = False):
        if(reload == True):
            p = Process(target=self.gernerate_route_data_process, args=(self.workspace, 
                                                                    json_path,
                                                                    self.flow,
                                                                    reload,
                                                                    self.input_def))
            p.start()
            p.join()
        else:
            ieda_feature = IEDAFeature(self.dir_workspace)
            # read json 
            ieda_feature.feature_route_read(json_path)
            
            # save route data to json
            ieda_feature.feature_route(json_path)
    
    # run as process
    def gernerate_route_data_process(self, workspace_path : WorkspacePath, 
                                 json_path : str,
                                 flow : DbFlow,
                                 reload : bool = False,
                                 input_def : str = ""):   
        # route data            
        ieda_feature = IEDAFeature(workspace_path.workspace)
        # if reload = True, read def again
        if(reload == True):
            if(input_def == None):
                input_def = workspace_path.get_output_def(flow)
            ieda_feature.read_def(input_def)
            
        # read json 
        ieda_feature.feature_route_read(json_path)
        
        # save route data to json
        ieda_feature.feature_route(json_path)

    def generate_eval_ieda(self):        
        ieda_feature = IEDAFeature(self.dir_workspace)
        eval_path = ieda_feature.feature_eval_summary(flow = self.flow)
        return eval_path
    
    def generate_timing_eval_ieda(self):        
        ieda_feature = IEDAFeature(self.dir_workspace)
        eval_path = ieda_feature.feature_timing_eval_summary(flow = self.flow)
        return eval_path
    
    def generate_net_eval_ieda(self):
        ieda_feature = IEDAFeature(self.dir_workspace)
        eval_path = ieda_feature.feature_net_eval(flow = self.flow)
        return eval_path
    
        


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   timing.py
@Time    :   2024/10/10 10:28:18
@Author  :   Dawn Li 
@Version :   1.0
@Contact :   dawnli619215645@gmail.com
@Desc    :   Timing Feature from Timing Eval iEDA
'''


from dataclasses import dataclass


@dataclass
class FeatureTiming(object):
    WNS: float = None
    TNS: float = None
    suggest_freq: float = None
    dynamic_power: float = None
    static_power: float = None



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   optimization.py
@Time    :   2024/08/06 12:44:09
@Author  :   SivanLaai
@Version :   1.0
@Contact :   laaisivan@gmail.com
@Desc    :   different optimization method
'''
import sys
import os
from multiprocessing import  Process
# set EDA tools working environment
# option : iEDA innovus PT dreamPlace
# if "eda_tool" not in os.environ:
#     os.environ['eda_tool'] = "iEDA"
current_dir = os.path.split(os.path.abspath(__file__))[0]
tool_dir = current_dir.rsplit('/', 2)[0]
sys.path.append(tool_dir)
from abc import abstractmethod, ABCMeta
from database.enum import EdaTool, FlowStep, FeatureOption, DSEMethod
from flow.flow_db import DbFlow
from engine.placement import EnginePlacement
from engine.routing import EngineRouting
from engine.net_opt import EngineNetOpt
from engine.timing_opt import EngineTimingOpt
from engine.cts import EngineCTS
from engine.gds import EngineGDS
# from flow.run_flow_ieda import RunFlowIEDA
from application.DSE.parameter import iEDAParameter, InnovusParameter, DREAMPlaceParameter
import wandb
import nni
import optuna
import numpy as np
from feature.io import FeatureIO
import time
import logging
from workspace.path import WorkspacePath
from application.DSE.arguements import Arguements
from application.DSE.config_management import ConfigManagement
import json

class AbstractOptimizationMethod(metaclass=ABCMeta):
    _parameter = None
    _search_config = None
    def __init__(self, args, workspace, parameter, algorithm="TPE", goal="minimize", step=FlowStep.place):
        self._method = algorithm
        self._goal = goal
        self._args = args
        self._workspace = workspace
        self._parameter = parameter
        self._step = step
        self._best_metric = float("inf")
        self.initOptimization()
    
    def getFeatureMetrics(self, data, eda_tool=EdaTool.IEDA, step=FlowStep.place, option=None):
        # evaluation
        eval_io = self.getFeatureIO(eda_tool, step, option=FeatureOption.eval)

        data["evaluation"] = eval_io.get_raw_json()

        place_data = dict()
        feature = self.getFeatureDB()
        hpwl = feature.wirelength.HPWL
        place_data["hpwl"] = hpwl
        if feature.timing:
            wns = feature.timing.FLUTE.clock_timings[0].hold_wns
            place_data["wns"] = wns
            tns = feature.timing.FLUTE.clock_timings[0].hold_tns
            place_data["tns"] = tns
        if len(place_data):
            data["place"] = place_data

        route_feature = self.getFeatureDB(eda_tool=EdaTool.IEDA, step=FlowStep.route, option=FeatureOption.tools)
        # # summary
        # tools_db = self.getFeatureDB(eda_tool, step, option=FeatureOption.tools)
        if route_feature.routing_summary:
            # route_wns = 0
            # route_tns = 0
            # route_freq = 0
            clock = route_feature.routing_summary.la_summary.clocks_timing[0]
            route_wns = clock.setup_wns
            route_tns = clock.setup_tns
            route_freq = clock.suggest_freq
            route_data = dict()
            route_data["route_tns"] = route_tns
            route_data["route_wns"] = route_wns
            route_data["route_freq"] = route_freq
            data["route"] = route_data
        return data

    def getFeatureIO(self, eda_tool=EdaTool.IEDA, step=FlowStep.place, option=FeatureOption.eval):
        io = FeatureIO(dir_workspace = self._workspace,
             eda_tool = eda_tool, 
             feature_option = option,
             flow = DbFlow(eda_tool = eda_tool,step = step))
        io.generate(reload = True)
        return io
    
    def getFeatureDB(self, eda_tool=EdaTool.IEDA, step=FlowStep.place, option=FeatureOption.eval):
        feature = FeatureIO(dir_workspace = self._workspace,
             eda_tool = eda_tool, 
             feature_option = option,
             flow = DbFlow(eda_tool = eda_tool,step = step))
        feature.generate(reload = True)
        db = feature.get_db()
        return db
    
    @abstractmethod
    def logFeature(self, metrics, step):
        raise NotImplementedError

    def setParameter(self, Parameter):
        self._parameter = Parameter
    
    def initOptimization(self):
        self._parameter._search_space = self._parameter.getSearchSpace()
        self.formatSweepConfig()

    @abstractmethod
    def formatSweepConfig(self):
        raise NotImplementedError
    
    @abstractmethod
    def loadParams(self, Parameter):
        raise NotImplementedError

    @abstractmethod
    def runOptimization(self, step=FlowStep.place, option=FeatureOption.tools, metrics={"hpwl": 1.0, "tns": 0.0, "wns": 0.0}, pre_step=FlowStep.cts, tool=EdaTool.IEDA):
        raise NotImplementedError

    @abstractmethod 
    def getNextParams(self):
       raise NotImplementedError
    
    def getPlaceResults(self):
        hpwl,wns,tns,freq = None,None,None,None
        try:
            workspace = self._workspace
            output_dir = os.path.join(workspace, "output/iEDA/data/pl/report/summary_report.txt")
            out_lines = open(output_dir).readlines()
            for i in range(len(out_lines)):
                line = out_lines[i]
                if "Total HPWL" in line:
                    hpwl = line.replace(" ", "").split("|")[-2]
                elif "Late TNS".lower() in line.lower() and "|" in out_lines[i+2]:
                    # print(line)
                    new_line = out_lines[i+2]
                    datas = new_line.replace(" ", "").split("|")
                    # print(datas)
                    wns = datas[-3]
                    tns = datas[-2]
            freq_dir = os.path.join(workspace, "output/iEDA/data/pl/log/info_ipl_glog.INFO") 
            # log_dir = os.path.join(workspace, "output/iEDA/data/pl/log") 
            # curr_time = datetime.now().strftime("%Y%m%d")
            # for file_path in os.listdir(log_dir):
            #     if "info_" in file_path and curr_time in file_path:
            # freq_dir = os.path.join(log_dir, file_path)
            lines = open(freq_dir).readlines()
            for i in range(len(lines)):
                line = lines[i]
                if "Freq(MHz)" in line and "|" in lines[i+2]:
                    next_line = lines[i+2]
                    # print(line)
                    datas = next_line.strip().replace(" ", "").split("|")
                    # print(datas)
                    freq = datas[-2]
                    if wns is None:
                        pass
                    if tns is None:
                        pass
                    break
        except Exception as e:
            print(e)
        return float(hpwl),float(wns),float(tns),float(freq)
    
    def getOperationEngine(self, step, tool, pre_step):
        dir_workspace = self._workspace
        project_name = self._args.project_name
        engine = None
        eda_tool = self.config_manage.getEdaTool()

        if step == FlowStep.floorplan:
            pass
        
        if step == FlowStep.fixFanout:
            pass
            
        if step == FlowStep.place:
            engine = EnginePlacement(dir_workspace = dir_workspace,
                             input_def=f"{dir_workspace}/output/iEDA/result/{project_name}_{pre_step.value}.def.gz",
                             input_verilog = f"{dir_workspace}/output/iEDA/result/{project_name}_{pre_step.value}.v.gz",
                             eda_tool = eda_tool,
                             pre_step = DbFlow(eda_tool = eda_tool,step = pre_step),
                             step = DbFlow(eda_tool = eda_tool,step = step))
        
        if step == FlowStep.cts:
            engine = EngineCTS(dir_workspace = dir_workspace,
                             input_def=f"{dir_workspace}/output/iEDA/result/{project_name}_{pre_step.value}.def.gz",
                             input_verilog = f"{dir_workspace}/output/iEDA/result/{project_name}_{pre_step.value}.v.gz",
                             eda_tool = tool,
                             pre_step = DbFlow(eda_tool = eda_tool,step = pre_step),
                             step = DbFlow(eda_tool = eda_tool,step = step))
            
        if step == FlowStep.optDrv:
            return FlowStep.optDrv
            
        if step == FlowStep.optHold:
            return FlowStep.optHold
            
        if step == FlowStep.optSetup:
            return FlowStep.optSetup
            
        if step == FlowStep.legalization:
            return FlowStep.legalization
            
        if step == FlowStep.route:
            engine = EngineRouting(dir_workspace = dir_workspace,
                             input_def=f"{dir_workspace}/output/iEDA/result/{project_name}_{pre_step.value}.def.gz",
                             input_verilog = f"{dir_workspace}/output/iEDA/result/{project_name}_{pre_step.value}.v.gz",
                             eda_tool = tool,
                             pre_step = DbFlow(eda_tool = eda_tool,step = pre_step),
                             step = DbFlow(eda_tool = eda_tool,step = step))
            
        if step == FlowStep.filler:
            return FlowStep.filler
            
        if step == FlowStep.gds:
            return FlowStep.gds
        
        if step == FlowStep.drc:
            return FlowStep.drc

        return engine

        
class NNIOptimization(AbstractOptimizationMethod):
    _parameter = None
    _search_config = dict()

    def __init__(self, args, workspace, parameter, algorithm="TPE", goal="minimize", step=FlowStep.place):
        super().__init__(args, workspace, parameter, algorithm, goal, step)
    def getNextParams(self):
        return nni.get_next_parameter()
    
    def loadParams(self, Parameter):
        self._parameter = Parameter

    def initOptimization(self):
        super().initOptimization()
        # self.formatSweepConfig()

    def formatSweepConfig(self):
        nni_search_space = self._parameter._search_space
        for key in nni_search_space:
            param = nni_search_space[key]
            if "distribution" in param:
                self._search_config[key] = {
                        "_type": param["distribution"],
                        "_value": [param["min"], param["max"]],
                    }
            else:
                self._search_config[key] = {
                        "_type": "choice",
                        "_value": param["values"],
                    }

    def logPlaceMetrics(self, metrics, results):
        feature = self.getFeatureDB()
        hpwl,wns,tns,freq = self.getPlaceResults()
        messages = ""
        metric = 0.0
        if hpwl is None:
            hpwl = feature.wirelength.HPWL
            wns = feature.timing.FLUTE.clock_timings[0].hold_wns
            tns = feature.timing.FLUTE.clock_timings[0].hold_tns

        hpwl_ref = metrics.get("hpwl", 1.0)
        messages += f"hpwl: {hpwl}, "
        metric += hpwl/hpwl_ref

        messages += f"wns: {wns}, "
        wns_ref = metrics.get("wns", 0.0)
        metric += np.exp(wns_ref)/np.exp(wns)

        messages += f"tns: {tns}, "
        tns_ref = metrics.get("tns", 0.0)
        metric += np.exp(tns_ref)/np.exp(tns)
        
        results["place_hpwl"] = hpwl
        results["place_wns"] = wns
        results["place_tns"] = tns
        results["place_freq"] = freq
        messages += f"place_hpwl: {hpwl}, place_wns: {wns}, place_tns: {tns}, place_freq: {freq}\n"
        logging.info(messages)
        return metric

    def logRouteMetrics(self, metrics, results):
        feature = self.getFeatureDB(eda_tool=EdaTool.IEDA, step=FlowStep.route, option=FeatureOption.tools)
        messages = ""
        metric = 0.0
        if feature.routing_summary:
            if feature.routing_summary.dr_summary:
                route_data = feature.routing_summary.dr_summary.summary[-1][-1]
                print(route_data)
                route_wl = route_data.total_wire_length
                clock = route_data.clocks_timing[-1]
                route_wns = clock.setup_wns
                route_tns = clock.setup_tns
                route_freq = clock.suggest_freq
                messages += f"route_wl: {route_wl}, route_tns: {route_tns}, route_wns: {route_wns}, route_freq: {route_freq}."
                # print(f"route_wl: {route_wl}, route_tns: {route_tns}, route_wns: {route_wns}, route_freq: {route_freq}.")
                metric += route_freq
                results["route_wl"] = route_wl
                results["route_tns"] = route_tns
                results["route_wns"] = route_wns
                results["route_freq"] = route_freq
                if "route_wl" in metrics:
                    metric += route_wl/metrics["route_wl"]
                if "route_tns" in metrics:
                    metric += np.exp(metrics["route_tns"])/np.exp(route_tns)
                if "route_wns" in metrics:
                    metric += np.exp(metrics["route_wns"])/np.exp(route_wns)
                if "route_freq" in metrics:
                    metric += route_freq/metrics["route_freq"]
        logging.info(messages)
        return metric
        
    def logFeature(self, metrics, step):
        metric = 0.0
        results = dict()
          
        if step == FlowStep.place:
            metric = self.logPlaceMetrics(metrics, results)
        else:
            self.logPlaceMetrics(metrics, results)
            metric = self.logRouteMetrics(metrics, results)
        if metric < self._best_metric:
            flow_list = self.config_manage.getFlowList()
            best_config_paths = self.config_manage.getBestConfigPathList()
            self._parameter.dumpOriginalConfig(best_config_paths, flow_list)
        nni.report_final_result(metric)

    def GenerateDataset(self, params, step=FlowStep.place, tool=EdaTool.IEDA):
        # if step==FlowStep.place:
        data = dict()
        data["params"] = params
        data = self.getFeatureMetrics(data, eda_tool=EdaTool.IEDA, step=FlowStep.place, option=None)
        # print(data)
        filepath = f"{self._args.result_dir}/benchmark/{self._args.tech}"
        filename = f"{self._args.result_dir}/benchmark/{self._args.tech}/{self._args.project_name}_{self._step.value}.jsonl"
        if not os.path.exists(filepath):
            os.makedirs(filepath)
        # print(filename)
        with open(filename, "a+") as bf:
            bf.write(json.dumps(data))
            bf.write("\n")
            bf.flush()
    
    def runTask(self, algorithm="TPE", goal="minimize", step=FlowStep.place, tool=EdaTool.IEDA, pre_step=FlowStep.cts):
        dir_workspace = self._workspace
        project_name = self._args.project_name
        engine = self.getOperationEngine(step, tool, pre_step)
        engine.run() 

    def runOptimization(self, step=FlowStep.place, option=FeatureOption.tools, metrics={"hpwl": 1.0, "tns": 0.0, "wns": 0.0}, pre_step=FlowStep.cts, tool=EdaTool.IEDA):
        tt = time.time()
        next_params = self.getNextParams()
        self.config_manage = ConfigManagement(self._args, tool)
        config_paths = self.config_manage.getConfigPathList()
        self._parameter.updateParams(next_params)
        flow_list = self.config_manage.getFlowList()
        self._parameter.dumpOriginalConfig(config_paths, flow_list)
        for db_flow in flow_list:
            step = db_flow.step
            pre_step = db_flow.pre_flow.step
            tool = db_flow.eda_tool
            p = Process(target=self.runTask, args=("bayes", "minimize", step, tool, pre_step)) #实例化进程对象
            p.start()
            p.join()
        self.logFeature(metrics, step)
        self.GenerateDataset(next_params, step, tool)
        total_time = time.time() - tt
        logging.info("task takes %.3f seconds" % (total_time))

class WandbOptimization(AbstractOptimizationMethod):
    _search_config = {
        'method': 'bayes',
        'metric': 
        {
            'goal': 'minimize', 
            'name': 'metric'
        },
        'parameters': None
    }
    def __init__(self, args, workspace, parameter, algorithm="bayes", goal="minimize", step=FlowStep.place):
        super().__init__(args, workspace, parameter, algorithm, goal, step)

    def initOptimization(self):
        # self.formatSweepConfig()
        super().initOptimization()
        sweep_id = wandb.sweep(self._search_config)
        wandb.agent(sweep_id, function=self.runOptimization, count=self._args.run_count)

    def formatSweepConfig(self):
        self._search_config["method"] = self._method 
        self._search_config["metric"]["goal"] = self._goal
        self._search_config["parameters"] = self._parameter._search_space

    def getNextParams(self):
        wandb.init(settings=wandb.Settings(start_method="fork"))
        return wandb.config
    
    def loadParams(self, Parameter):
        self._parameter = Parameter

    def GenerateDataset(self, params, step=FlowStep.place, tool=EdaTool.IEDA):
        # if step==FlowStep.place:
        summary = feature.place_summary
        params["hpwl"] = place.gplace.HPWL 
        params["hpwl"] = place.gplace.wns
        params["hpwl"] = place.gplace.tns
        filepath = f"{self.args.result_dir}/benchmark/{self.args.tech}/{self.args.project_name}.jsonl"
        print(filepath)
        with open(filepath, "a+") as bf:
            bf.write(json.dumps(params))
            bf.write("\n")
            bf.flush()
    
    def runTask(self, algorithm="bayes", goal="minimize", step=FlowStep.place, tool=EdaTool.IEDA, pre_step=FlowStep.cts):
        dir_workspace = self._workspace
        eda_tool = self.config_manage.getEdaTool()
        if step == FlowStep.full_flow:
            flow = RunFlowIEDA(dir_workspace = dir_workspace)
            result = flow.run()
            flow.reset_flow_state()
        else:
            engine = EnginePlacement(dir_workspace = dir_workspace,
                 input_def=f"{dir_workspace}/output/iEDA/result/gcd_legalization.def",
                 input_verilog = f"{dir_workspace}/output/iEDA/result/gcd_legalization.v",
                 eda_tool = eda_tool,
                 pre_step = DbFlow(eda_tool = eda_tool,step = pre_step),
                 step = DbFlow(eda_tool = eda_tool,step = step))
            engine.run()
    
    def logFeature(self, metrics, step):
        feature = self.getFeatureDB()
        place = feature.place_summary
        metric = 0.0
        if "hpwl" in metrics:
            hpwl = place.gplace.HPWL
            wandb.log({"hpwl": hpwl})
            hpwl_ref = metrics["hpwl"]
            metric += hpwl/hpwl_ref
        if "wns" in metrics:
            wns = place.gplace.wns
            wandb.log({"wns": wns})
            wns_ref = metrics["wns"]
            metric += np.exp(wns_ref)/np.exp(wns)
        if "tns" in metrics:
            tns = place.gplace.tns
            wandb.log({"tns": tns})
            tns_ref = metrics["tns"]
            metric += np.exp(tns_ref)/np.exp(tns)
        metric /= len(metrics.keys())
        wandb.log({"metric": metric})
    
    
    def runOptimization(self, step=FlowStep.place, option=FeatureOption.tools, metrics={}, pre_step=FlowStep.cts, tool=EdaTool.IEDA):
        # flow
        tt = time.time()
        next_params = self.getNextParams()
        self._parameter.updateParams(next_params)

        config_manage = ConfigManagement(self.args, eda_tool)
        config_path = config_manage.getConfigPath()
        flow_list = config_manage.getFlowList()
        self._parameter.dumpOriginalConfig(config_path, step, flow_list)
        for db_flow in flow_list:
            step = db_flow.step
            pre_step = db_flow.pre_flow.step
            tool = db_flow.eda_tool
            self.engine = config_manage.getEngine(step)
            p = Process(target=self.runTask, args=("bayes", "minimize", step, tool, pre_step)) #实例化进程对象
            p.start()
            p.join()
        # self.runTask()
        self.logFeature(feature, metrics, step)
        total_time = time.time() - tt
        logging.info("task takes %.3f seconds" % (total_time))
        wandb.log({"total_time": total_time})

if __name__ == "__main__":
    
    tool = os.environ.get('eda_tool', "iEDA")
    args = Arguements.parse(sys.argv[1:])
    parameter = os.environ.get('eda_tool', "iEDA")
    config_manage = ConfigManagement(args, tool)
    step = config_manage.getStep()
    dir_workspace = config_manage.getWorkspacePath()
    eda_tool = config_manage.getEdaTool()

    params = config_manage.getParameters()

    method = NNIOptimization(args, dir_workspace, params, algorithm="TPE", goal="maximize", step=step)
    method.runOptimization(tool=eda_tool, step=step, pre_step=FlowStep.place, metrics={})
    # method.runOptimization(tool=eda_tool, step=step, pre_step=FlowStep.place)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : run_pr_flow.py
@Author : yell
@Desc : run all flow 
'''
import os

from flow.flow_db import DbFlow
from database.enum import EdaTool, FlowStep
from workspace.path import WorkspacePath
from flow.flow_manager import FlowManager
from tools.fpt.api.tech_api import TechDRCsApi
from engine.floorplan import EngineFloorplan
from engine.cts import EngineCTS
from engine.placement import EnginePlacement
from engine.routing import EngineRouting
from engine.drc import EngineDRC
from engine.full_flow import EngineFullFlow

class RunFlowOpenROAD:
    '''run eda backend flow
    '''
    
    def __init__(self, dir_workspace : str, dir_resource : str=""):
        self.dir_workspace = dir_workspace
        self.dir_resource = dir_resource
        
        self.workspace = WorkspacePath(dir_workspace, dir_resource)
        self.flow_manager = FlowManager(self.workspace.get_config('flow'))
    
    def run(self):
        """run flow"""  
        flow = self.flow_manager.get_new_flow()
        while flow is not None :
            #run flow 
            self.flow_manager.set_current_flow(flow)
            if self.run_flow(flow) is False:
                return False
            #set as pre flow
            self.flow_manager.set_pre_flow(flow)
            #get next flow
            flow = self.flow_manager.get_new_flow()
          
        #check all flow success  
        for check_flow in self.flow_manager.flow_list :
            if check_flow.is_finish() is False:
                return False
        
        return True
            
    def run_flow(self, flow : DbFlow):
        """run flow"""    
        #set state running
        flow.set_state_running()
        self.flow_manager.save_flow_state(flow)
            
        #run eda tool
        self.run_eda_tool(flow)
        
        #save flow state
        if self.check_flow_state(flow) is True:
            flow.set_state_finished()
            self.flow_manager.save_flow_state(flow)
            return True
        else:
            flow.set_state_imcomplete()   
            self.flow_manager.save_flow_state(flow)
            return False 
            
    def run_eda_tool(self,  flow : DbFlow):
        io_files = self.build_io_file(flow)

        """run eda tool"""
        
        if (flow.eda_tool == EdaTool.INNOVUS):
            # run innovus
            self.build_flow(flow.step, io_files)

            print(io_files)
        else:
            pass

    def build_io_file(self, flow : DbFlow):
        io_files = {}
        
        io_files["input_def"] = self.workspace.get_input_def(flow)
        io_files["input_verilog"] = self.workspace.get_input_verilog(flow)
            
        io_files["output_def"] = self.workspace.get_output_def(flow)
        io_files["output_verilog"] = self.workspace.get_output_verilog(flow)
        
        return io_files        
    
    def check_flow_state(self, flow : DbFlow):
        """check state"""
        #check flow success if output def & verilog file exist
        output_def = self.workspace.get_output_def(flow)
        output_verilog = self.workspace.get_output_verilog(flow)
        
        return (os.path.exists(output_def) and os.path.exists(output_verilog))
        
   
    def build_flow(self, step: FlowStep, io_files : dict):
        print("Current Step:", step)
        if(step == FlowStep.floorplan):
            run_flow = EngineFloorplan(dir_workspace = self.dir_workspace,
                                      eda_tool = EdaTool.OPENROAD,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            run_flow.run()
        
        if(step == FlowStep.prePlace):
            run_flow = EnginePlacement(dir_workspace = self.dir_workspace,
                                      eda_tool = EdaTool.OPENROAD,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            run_flow.run_pre_placer()
            
        if(step == FlowStep.place):
            run_flow = EnginePlacement(dir_workspace = self.dir_workspace,
                                      eda_tool = EdaTool.OPENROAD,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            run_flow.run_placer()
        
        if(step == FlowStep.cts):
            run_flow = EngineCTS(dir_workspace = self.dir_workspace,
                                      eda_tool = EdaTool.OPENROAD,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            run_flow.run()
            
        if(step == FlowStep.route):
            run_flow = EngineRouting(dir_workspace = self.dir_workspace,
                                      eda_tool = EdaTool.OPENROAD,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            run_flow.run()


#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
'''
@File    :   utility.py
@Time    :   2022/09/15 15:49:52
@Author  :   shuaiying long
@Version :   1.0
@Contact :   longshy@pcl.ac.cn
@Desc    :   The definitions of common functions.
'''

def find_all_char(string, ch):
    pos = []
    start_len = 0
    while True:
        index = string.find(ch, start_len)
        if index == -1:
            break
        pos.append(index)
        start_len = index+1
    return pos


def find_last_char(string, ch):
    pos = []
    start_len = 0
    while True:
        index = string.find(ch, start_len)
        if index == -1:
            break
        pos.append(index)
        start_len = index+1
    return pos[-1]





#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : sta.py
@Author : yell
@Desc : STA framework
'''
from engine.base import EngineBase
from database.enum import EdaTool
from flow.flow_db import DbFlow
from tools.iEDA.module.sta import IEDASta
from database.enum import FlowStep, EdaTool
from tools.innovus.module.sta import InnovusSTA
from dataclasses import dataclass
from typing import Optional

@dataclass(frozen=True)
class StaRCNode:
    """RC Tree Node"""
    coord_x : int
    coord_y : int
    layer : int
    
    node_id : int

    def __repr__(self):
        return f"StaRCNode(coord_x={self.coord_x}, coord_y={self.coord_y}, layer={self.layer})"

@dataclass(frozen=True)
class StaRCObjNode(StaRCNode):
    """RC Tree Object Node"""
    obj_node : str

@dataclass(frozen=True)
class StaRCEdge:
    """RC Tree Edge"""
    src_node : StaRCNode
    dst_node : StaRCNode

from dataclasses import dataclass, field

@dataclass    
class StaRCTree:
    """RC Tree"""
    nodes: set = field(default_factory=set)
    edges: list = field(default_factory=list)
    driver_node: Optional[StaRCObjNode] = None
    load_nodes: list = field(default_factory=list)
    
    def get_node_id(self, node):
        for i, n in enumerate(self.nodes):
            if n == node:
                return i
    def draw(self, filename=None):
        import matplotlib.pyplot as plt
        import networkx as nx

        G = nx.DiGraph()
        pos = {}

        # Add nodes
        for node in self.nodes:
            pos[node] = (node.coord_x, node.coord_y)
            G.add_node(node)

        # Add edges
        for edge in self.edges:
            G.add_edge(edge.src_node, edge.dst_node)

        # Draw all nodes
        nx.draw_networkx_nodes(G, pos, node_color='lightblue', node_size=300)
        # Draw edges
        nx.draw_networkx_edges(G, pos, arrows=True)

        # Highlight driver node in red
        if self.driver_node and self.driver_node in G.nodes:
            nx.draw_networkx_nodes(G, pos, nodelist=[self.driver_node], node_color='red', node_size=400)

        # Highlight load nodes in red
        load_nodes_in_graph = [n for n in self.load_nodes if n in G.nodes]
        if load_nodes_in_graph:
            nx.draw_networkx_nodes(G, pos, nodelist=load_nodes_in_graph, node_color='red', node_size=400)

        # Draw labels
        labels = {node: str(node) for node in G.nodes}
        nx.draw_networkx_labels(G, pos, labels, font_size=8)

        plt.axis('off')
        if filename:
            plt.savefig(filename, bbox_inches='tight')
        else:
            plt.show()
        plt.close()
        
    def graphviz_draw(self, filename=None):
        import graphviz
        dot = graphviz.Digraph(format='png')

        # Add nodes
        for node in self.nodes:
            label = str(node)
            if node == self.driver_node:
                dot.node(str(id(node)), label, color='red', style='filled', fillcolor='red')
            elif node in self.load_nodes:
                dot.node(str(id(node)), label, color='red', style='filled', fillcolor='red')
            else:
                dot.node(str(id(node)), label)

        # Add edges
        for index, edge in enumerate(self.edges):
            dot.edge(str(id(edge.src_node)), str(id(edge.dst_node)))
            
        with open("graphviz_output.dot", "w", encoding="utf-8") as f:
            f.write(dot.source)

        if filename:
            dot.render(filename, view=False, cleanup=True)
        else:
            from IPython.display import display
            display(dot)

@dataclass 
class StaDesignRCTree:
    """Design RC Tree"""
    rc_trees = {}

class EngineSTA(EngineBase):
    """STA framework"""      
    def run(self):
        return super().run()

    def run_ieda(self):
        # run iEDA
        run_flow = IEDASta(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_sta()

        return self.get_result()

    def init_sta(self):
        """init sta"""
        sta_engine = IEDASta(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        sta_engine.init_sta()

    def make_rc_tree(self, timing_json_file: str) -> StaDesignRCTree:
        import json

        design_rc_tree = StaDesignRCTree()

        with open(timing_json_file, 'r', encoding='utf-8') as f:
            for line in f:                
                data = json.loads(line)
                driver = data['driver']
                if driver not in design_rc_tree.rc_trees:
                    design_rc_tree.rc_trees[driver] = StaRCTree()
                else:
                    print("find exist driver %s" % driver)

                one_rc_tree = design_rc_tree.rc_trees[driver]

                load = data['load']
                points = data['points']                

                driver_node = StaRCObjNode(points[0][0], points[0][1], points[0][2], len(one_rc_tree.nodes), driver)
                if driver_node not in one_rc_tree.nodes:
                    one_rc_tree.driver_node = driver_node
                    one_rc_tree.nodes.add(one_rc_tree.driver_node)

                load_node = StaRCObjNode(points[-1][0], points[-1][1], points[-1][2], len(one_rc_tree.nodes), load)
                if load_node not in one_rc_tree.nodes:
                    one_rc_tree.load_nodes.append(load_node)
                    one_rc_tree.nodes.add(load_node)
                else:
                    continue

                previous_node = None
                for point in points:
                    if point == points[0]:
                        previous_node = one_rc_tree.driver_node
                        continue

                    if point == points[-1]:
                        if previous_node:
                            one_rc_tree.edges.append(StaRCEdge(previous_node, one_rc_tree.load_nodes[-1]))
                        continue

                    one_point_node = StaRCNode(point[0], point[1], point[2], len(one_rc_tree.nodes))
                    if one_point_node not in one_rc_tree.nodes:
                        one_rc_tree.nodes.add(one_point_node)
                        if previous_node:
                            one_rc_tree.edges.append(StaRCEdge(previous_node, one_point_node))

                        previous_node = one_point_node
                    else:
                        print("point node already exists %s" % one_point_node)
                        for point_node in one_rc_tree.nodes:
                            if hash(one_point_node) == hash(point_node):
                                previous_node = point_node

            print("build rc tree done.")

        return design_rc_tree

    def init_sta_one_rc(self, one_net_rc_tree: StaRCTree):
        sta_engine = IEDASta(dir_workspace = self.dir_workspace,
                        input_def = self.input_def, 
                        output_def = self.output_def, 
                        output_verilog = self.output_verilog)

        driver_node = one_net_rc_tree.driver_node
        if driver_node is not None:
            driver_name = driver_node.obj_node
        else:
            driver_name = None
            raise Exception("Driver node not found")

        net_name = sta_engine.get_net_name(driver_name)
        net_segment_edge = one_net_rc_tree.edges
        for edge in net_segment_edge:
            edge_src_node = edge.src_node
            edge_dst_node = edge.dst_node
            
            src_node_id = edge_src_node.node_id
            snk_node_id = edge_dst_node.node_id

            if edge_src_node.layer == edge_dst_node.layer:
                layer_id = (int)(edge_src_node.layer / 2) + 2  # layer id TBD
                dbu = 2000.0 # hard code dbu
                segment_length = abs(
                    edge_dst_node.coord_x - edge_src_node.coord_x
                ) + abs(edge_dst_node.coord_y - edge_src_node.coord_y)
                
                segment_length = segment_length / dbu

                edge_resistance = sta_engine.get_net_segment_resistance(
                    layer_id, segment_length
                )
                edge_capacitance = sta_engine.get_net_segment_capacitance(
                    layer_id, segment_length
                )
            else:
                # via tbd
                edge_capacitance = 0.001
                edge_resistance = 5
            
            if isinstance(edge_src_node, StaRCObjNode):
                src_rc_node_name = sta_engine.make_rc_tree_obj_node(
                    edge_src_node.obj_node, edge_capacitance / 2
                )

            else:
                src_rc_node_name = sta_engine.make_rc_tree_inner_node(
                    net_name, src_node_id, edge_capacitance / 2
                )
                
            if isinstance(edge_dst_node, StaRCObjNode):
                dst_rc_node_name = sta_engine.make_rc_tree_obj_node(
                    edge_dst_node.obj_node, edge_capacitance / 2
                )
            else:
                dst_rc_node_name = sta_engine.make_rc_tree_inner_node(
                    net_name, snk_node_id, edge_capacitance / 2
                )
                
            sta_engine.make_rc_tree_edge(
                    net_name, src_rc_node_name, dst_rc_node_name, edge_resistance
                )
        sta_engine.update_rc_tree_info(net_name)
        
    def init_sta_rc(self, design_rc_tree: StaDesignRCTree):
        for one_net_rc_tree in design_rc_tree.rc_trees.values():
            self.init_sta_one_rc(one_net_rc_tree)
        
    def update_and_report_sta(self):
        sta_engine = IEDASta(dir_workspace = self.dir_workspace,
                        input_def = self.input_def, 
                        output_def = self.output_def, 
                        output_verilog = self.output_verilog)
        
        sta_engine.update_timing()
        sta_engine.report_sta()
        return sta_engine.get_wire_timing_data()

    def run_innovus(self):
        if(self.step == None):
            self.step = DbFlow(eda_tool=EdaTool.INNOVUS,
                          pre_flow=self.pre_step,
                          step=FlowStep.sta)

        self.build_path()

        # run innovus
        run_flow = InnovusSTA(dir_workspace = self.dir_workspace,
                               dir_resource = self.dir_resource,
                              input_def = self.input_def, 
                              input_verilog = self.input_verilog,
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              pre_step = self.pre_step,
                              step = self.step,
                              task = self.task)
        run_flow.run_sta()

        return self.get_result()

    def refine_libs(self):
        run_flow = IEDASta(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)

        libs = run_flow.get_used_lib()

        # refine
        self.workspace.reset_path_libs(libs)
        self.workspace.reset_path_ieda_libs(libs)



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
'''
@File : drc.py
@Author : yell
@Desc : drc database
'''

import sys
import os

from dataclasses import dataclass
from dataclasses import field
from typing import List


@dataclass
class DrcShape(object):
    llx: float = None
    lly: float = None
    urx: float = None
    ury: float = None
    net_ids: list = field(default_factory=list)
    inst_ids: list = field(default_factory=list)


@dataclass
class DrcLayer(object):
    layer: str = None
    number: int = None
    shapes: List[DrcShape] = field(default_factory=list)


@dataclass
class DrcDistribution(object):
    type: str = None
    number: int = None
    layers: List[DrcLayer] = field(default_factory=list)


@dataclass
class DrcDistributions(object):
    number: int = None
    drc_list: List[DrcDistribution] = field(default_factory=list)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   timing.py
@Time    :   2024/10/10 14:54:56
@Author  :   Dawn Li
@Version :   1.0
@Contact :   dawnli619215645@gmail.com
@Desc    :   Timing Feature Parser
'''


from typing import Dict
import json
import os
import sys
import logging

if True:
    current_dir = os.path.split(os.path.abspath(__file__))[0]
    tool_dir = current_dir.rsplit('/', 1)[0]
    sys.path.append(tool_dir)

    from tools.innovus.feature.database.tools import PowerINNOVUSBase, TimingINNOVUSBase, FeatureTimingINNOVUS
    from tools.iEDA.feature.database.eval import FeatureTimingIEDA, MethodTimingIEDA, FeatureTimingEnumIEDA, FeatureEval
    from tools.iEDA.feature.database.tools import ClockTiming
    from database.timing import FeatureTiming
    from database.enum import EdaTool, FeatureOption
    from flow.flow_db import DbFlow
    from feature.io import FeatureIO, FeatureOption, FlowStep
    from workspace.path import WorkspacePath
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s] %(message)s',
    handlers=[
        logging.StreamHandler()
    ]
)
logging.basicConfig(
    level=logging.WARNING,
    format=f'%(asctime)s [\033[95m%(levelname)s\033[0m] %(message)s',
    handlers=[
        logging.StreamHandler()
    ]
)


class FeatureTimingBase(object):
    def __init__(self, dir_workspace: str, eda_tool: EdaTool, step: FlowStep):
        self.dir_workspace = dir_workspace
        self.eda_tool = eda_tool
        self.step = step
        self.workspace = WorkspacePath(dir_workspace)


class FeatureTimingGenerator(FeatureTimingBase):
    def __init__(self, dir_workspace: str, eda_tool: EdaTool, step: FlowStep):
        super().__init__(dir_workspace, eda_tool, step)

    def generateFeatureTimingIEDA(self, eval_path: str) -> FeatureTimingIEDA:
        """generate timing feature from timing eval ieda

        Args:
            eval_path (str): path of timing eval ieda

        Returns:
            Timing: Timing Feature
        """
        if not os.path.exists(eval_path):
            logging.warning(f"{eval_path} not found")
            return None
        # Initialize MethodTiming instances for each method
        method_timing_dict: Dict[str, MethodTimingIEDA] = {}
        for method in FeatureTimingEnumIEDA:
            method_timing_dict[method.value] = MethodTimingIEDA([], 0, 0)

        with open(eval_path, 'r') as file:
            for line in file:
                data = json.loads(line.strip())
                if 'Timing' in data:
                    timing_data = data['Timing']
                    for method_name, timing_entries in timing_data.items():
                        # Check if method_name is one of the expected methods
                        if method_name in method_timing_dict:
                            method_timing = method_timing_dict[method_name]
                            for timing_entry in timing_entries:
                                # Create a TimingData instance for each clock
                                timing_data_obj = ClockTiming(
                                    clock_name=timing_entry['clock_name'],
                                    hold_tns=timing_entry['hold_tns'],
                                    hold_wns=timing_entry['hold_wns'],
                                    setup_tns=timing_entry['setup_tns'],
                                    setup_wns=timing_entry['setup_wns'],
                                    suggest_freq=timing_entry['suggest_freq']
                                )
                                method_timing.clock_timings.append(
                                    timing_data_obj)
                elif 'Power' in data:
                    power_data = data['Power']
                    for method_name, power_entry in power_data.items():
                        # Check if method_name is one of the expected methods
                        if method_name in method_timing_dict:
                            method_timing = method_timing_dict[method_name]
                            method_timing.dynamic_power = power_entry['dynamic_power']
                            method_timing.static_power = power_entry['static_power']

        # Create the Timing instance with the collected data
        timing_ieda = FeatureTimingIEDA()
        for method_name, method_timing in method_timing_dict.items():
            setattr(timing_ieda, method_name, method_timing)

        return timing_ieda

    def generateFeatureTimingINNOVUS(self, sta_path, power_path) -> FeatureTimingINNOVUS:
        timing = TimingINNOVUSBase()
        power = PowerINNOVUSBase()

        if not os.path.exists(sta_path):
            print(f"\033[35mWarning: {sta_path} not found\033[0m")
            return None
        if not os.path.exists(power_path):
            print(f"\033[35mWarning: {power_path} not found\033[0m")
            return None

        with open(sta_path, 'r') as file:
            data = json.load(file)
            timing.WNS = data['sta']['WNS']
            timing.TNS = data['sta']['TNS']
            timing.violation_paths = data['sta']['Violating Paths']
            timing.max_cap = data['sta']['max_cap']
            timing.max_tran = data['sta']['max_tran']
            timing.max_fanout = data['sta']['max_fanout']
            timing.max_length = data['sta']['max_length']
            timing.density = data['sta']['Density']
            timing.frequency = data['frequency']

        with open(power_path, 'r') as file:
            data = json.load(file)
            factor = 1e-3  # convert mW to W
            power.internal_power = data['power']['internal_power'] * factor
            power.switch_power = data['power']['switch_power'] * factor
            power.leakage_power = data['power']['leakage_power'] * factor
            power.total_power = data['power']['total_power'] * factor

        timing_innovus = FeatureTimingINNOVUS(timing=timing, power=power)
        return timing_innovus

    def convertFeatureTimingIEDA(self, timing_ieda: FeatureTimingIEDA, method: FeatureTimingEnumIEDA = FeatureTimingEnumIEDA.FLUTE) -> FeatureTiming:
        timing_feature = FeatureTiming()
        if timing_ieda is None:
            logging.warning(
                f"timing_ieda is None, dir_workspace: {self.dir_workspace}")
            return timing_feature
        method_timing: MethodTimingIEDA = getattr(timing_ieda, method.value)
        clock_timing: ClockTiming = method_timing.clock_timings[0]
        timing_feature.WNS = clock_timing.setup_wns
        timing_feature.TNS = clock_timing.setup_tns
        timing_feature.suggest_freq = clock_timing.suggest_freq
        timing_feature.dynamic_power = method_timing.dynamic_power
        timing_feature.static_power = method_timing.static_power
        return timing_feature

    def convertFeatureTimingINNOVUS(self, timing_innovus: FeatureTimingINNOVUS) -> FeatureTiming:
        timing_feature = FeatureTiming()
        if timing_innovus is None:
            logging.warning(
                f"timing_innovus is None, dir_workspace: {self.dir_workspace}")
            return timing_feature
        if timing_innovus.timing is None:
            logging.warning(
                f"timing_innovus.timing is None, dir_workspace: {self.dir_workspace}")
            return timing_feature
        timing_feature.WNS = timing_innovus.timing.WNS
        timing_feature.TNS = timing_innovus.timing.TNS
        timing_feature.suggest_freq = timing_innovus.timing.frequency
        if timing_innovus.power is None:
            logging.warning(
                f"timing_innovus.power is None, dir_workspace: {self.dir_workspace}")
            return timing_feature
        timing_feature.dynamic_power = timing_innovus.power.internal_power + \
            timing_innovus.power.switch_power
        timing_feature.static_power = timing_innovus.power.leakage_power
        return timing_feature

    def getFeatureTiming(self) -> FeatureTiming:
        if self.eda_tool == EdaTool.IEDA:
            return self.getFeatureTimingFromIEDA()
        elif self.eda_tool == EdaTool.INNOVUS:
            return self.getFeatureTimingFromINNOVUS()

    def getFeatureTimingFromIEDA(self) -> FeatureTiming:
        feature_io = FeatureIO(dir_workspace=self.dir_workspace,
                               eda_tool=self.eda_tool,
                               feature_option=FeatureOption.eval,
                               flow=DbFlow(eda_tool=self.eda_tool, step=self.step))
        feature_db: FeatureEval = feature_io.get_db_ieda()
        feature = self.convertFeatureTimingIEDA(feature_db)
        return feature

    def getFeatureTimingFromINNOVUS(self) -> FeatureTiming:
        if(self.eda_tool == EdaTool.IEDA):
            # HARD CODE: eda_tool=EdaTool.IEDA
            timing_feature_io = FeatureIO(dir_workspace=self.dir_workspace,
                                          eda_tool=self.eda_tool,
                                          feature_option=FeatureOption.baseline_sta,
                                          flow=DbFlow(eda_tool=self.eda_tool, step=self.step))
            timing_db: TimingINNOVUSBase = timing_feature_io.get_db_ieda()
    
            # HARD CODE: eda_tool=EdaTool.IEDA
            power_feature_io = FeatureIO(dir_workspace=self.dir_workspace,
                                         eda_tool=self.eda_tool,
                                         feature_option=FeatureOption.baseline_power,
                                         flow=DbFlow(eda_tool=self.eda_tool, step=self.step))
            power_db: PowerINNOVUSBase = power_feature_io.get_db_ieda()
    
            feature_db = FeatureTimingINNOVUS(timing=timing_db, power=power_db)
            feature = self.convertFeatureTimingINNOVUS(feature_db)
            return feature
        
        if(self.eda_tool == EdaTool.INNOVUS):
            timing_feature_io = FeatureIO(dir_workspace=self.dir_workspace,
                                          eda_tool=self.eda_tool,
                                          feature_option=FeatureOption.tools,
                                          flow=DbFlow(eda_tool=self.eda_tool, step=self.step))
            
            timing_db = None
            power_db = None

            if(self.step == FlowStep.place):
                timing_db, power_db, db_hotspot = timing_feature_io.get_db_innovus()
            if(self.step == FlowStep.cts):
                timing_db, power_db = timing_feature_io.get_db_innovus()
            if(self.step == FlowStep.route):
                timing_db, power_db = timing_feature_io.get_db_innovus()

            feature_db = FeatureTimingINNOVUS(timing=timing_db, power=power_db)
            feature = self.convertFeatureTimingINNOVUS(feature_db)
            
            return feature
        
        return None


if __name__ == "__main__":

    def get_ieda_eval_path(case_name, step):
        return os.path.join("/data/project_share/dataset_baseline/", case_name, "workspace/output/iEDA/feature/", f"{case_name}_{step}_eval.jsonl")

    def get_innovus_power_path(case_name, step):
        return os.path.join("/data/project_share/dataset_baseline/", case_name, "workspace/output/iEDA/feature/", f"{case_name}_{step}_baseline_power.json")

    def get_innovus_sta_path(case_name, step):
        return os.path.join("/data/project_share/dataset_baseline/", case_name, "workspace/output/iEDA/feature/", f"{case_name}_{step}_baseline_sta.json")

    def get_workspace_path(case_name):
        return os.path.join("/data/project_share/dataset_baseline/", case_name, "workspace")

    test_case = "aes"
    test_step = "place"
    feature_parser = FeatureTimingGenerator(
        dir_workspace=get_workspace_path(test_case), eda_tool=EdaTool.IEDA, step=FlowStep.place)

    ieda_timing_feature: FeatureTimingIEDA = feature_parser.generateFeatureTimingIEDA(
        get_ieda_eval_path(test_case, test_step))

    logging.info(f"iEDA Feature:\n{ieda_timing_feature}")

    innovus_timing_feature = feature_parser.generateFeatureTimingINNOVUS(
        get_innovus_sta_path(test_case, test_step), get_innovus_power_path(test_case, test_step))

    logging.info(f"INNOVUS Feature:\n{innovus_timing_feature}")

    feature_ieda = feature_parser.convertFeatureTimingIEDA(
        ieda_timing_feature)

    logging.info(f"Timing Feature built from iEDA:\n{feature_ieda}")

    feature_innovus = feature_parser.convertFeatureTimingINNOVUS(
        innovus_timing_feature)

    logging.info(f"Timing Feature built from INNOVUS:\n{feature_innovus}")




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   parameter.py
@Time    :   2024/08/06 12:44:09
@Author  :   SivanLaai
@Version :   1.0
@Contact :   laaisivan@gmail.com
@Desc    :   parameter space operation for different engines
'''
import sys
import os
# set EDA tools working environment
# option : iEDA innovus PT dreamPlace

current_dir = os.path.split(os.path.abspath(__file__))[0]
tool_dir = current_dir.rsplit('/', 2)[0]
sys.path.append(tool_dir)
from abc import abstractmethod, ABCMeta
import json 
import math 
from collections import OrderedDict
import logging
from database.enum import EdaTool, FlowStep 

class AbstractParameter(metaclass=ABCMeta):
    _search_space = None
    config = {}
    next_params = {}

    '''
    @Func 
    @Desc return
    @Param filename | default parameter json path
           step | step of AiEDA flow
    @Return None
    '''
    def __init__(self, filename="./config/iEDA/default.json", step=FlowStep.place):
        """
        @brief initialization
        """
        filename = os.path.join(os.path.dirname(__file__), filename)
        self._step = step
        self.__dict__ = {}
        # print(step)
        # print(self._step)
        self.initData(filename, step)
        self.formatSearchSpace(step)

    @abstractmethod
    def dumpPlaceFlowConfig(self, filename, config=None, step=FlowStep.place):
        raise NotImplementedError

    @abstractmethod
    def dumpCTSFlowConfig(self, filename, config=None, step=FlowStep.cts):
        raise NotImplementedError

    @abstractmethod
    def dumpRouteFlowConfig(self, filename, config=None, step=FlowStep.route):
        raise NotImplementedError
    
    @abstractmethod
    def dumpFullFlowConfig(self, filename, config=None, flow_list=None):
        raise NotImplementedError

    def dumpConfigByFlowStep(self, filename, flow_step):
        if flow_step == FlowStep.place:
            # self.params_dict[flow_step.value.lower()]
            self.dumpPlaceFlowConfig(filename)
        elif flow_step == FlowStep.cts:
            self.dumpCTSFlowConfig(filename)
        else:
            pass

    @abstractmethod
    def formatSearchSpace(self, step=None):
        raise NotImplementedError

    def getSearchSpace(self, step=None):
        return self._search_space

    @abstractmethod
    def dumpOriginalConfig(self, filename, step=None, flow_list=None):
        raise NotImplementedError
    
    def initData(self, filename, step):
        # print(step)
        params_dict = {}
        with open(filename, "r") as f:
            params_dict = json.load(f, object_pairs_hook=OrderedDict)
        self.__dict__ = dict()
        for param_step in params_dict:
            step_params_dict = params_dict[param_step]
            for key, value in step_params_dict.items():
                if 'default' in value: 
                    if isinstance(value['default'], dict):
                        self.__dict__[key] = dict()
                        for k,v in value['default'].items():
                            self.__dict__[key][k] = v
                    else:
                        self.__dict__[key] = value['default']
                else:
                    self.__dict__[key] = None
        self.__dict__['params_dict'] = params_dict
    
    def getParamsDict(self):
        return self.params_dict
    
    def get(self, key):
        return self.__dict__.get(key, None)

    def set(self, key, value):
        if "." not in key:
            self.__dict__[key] = value
        else:
            key1, key2 = key.split(".")
            if key1 in self.__dict__:
                if isinstance(self.__dict__[key1] ,list):
                    if key2 in self.__dict__[key1][0]:
                        self.__dict__[key1][0][key2] = value
                elif key2 in self.__dict__[key1]:
                    self.__dict__[key1][key2] = value
        if "." not in key:
            self.config[key] = value
        else:
            key1, key2 = key.split(".")
            if key1 in self.config:
                if isinstance(self.config[key1] ,list):
                    if key2 in self.config[key1][0]:
                        self.config[key1][0][key2] = value
                elif key2 in self.config[key1]:
                    self.config[key1][key2] = value

    def toJson(self):
        """
        @brief convert to json
        """
        data = {}
        for key, value in self.__dict__.items():
            if key != 'params_dict': 
                data[key] = value
        return data

    def fromJson(self, data):
        """
        @brief load form json
        """
        for key, value in data.items(): 
            self.__dict__[key] = value


    def dump(self, filename):
        """
        @brief dump to json file
        """
        with open(filename, 'w') as f:
            json.dump(self.toJson(), f)

    def load_list(self, path_list):
        """
        @brief load from json file list
        """
        for step in path_list:
            filepath = path_list[step]
            self.load(filepath, step)
        print(self.config, "load_list")
        self.fromJson(self.config)

    def load(self, filename, step=None):
        """
        @brief load from json file
        """
        with open(filename, 'r') as f:
            data = json.load(f)
            if step is not None:
                self.config[step] = data
            # self.fromJson(self.config[step])

    def updateParams(self, new_param_dict):
        self.next_params = new_param_dict
        print("update self.next_params:", self.next_params)
        # for key,value in new_param_dict.items():
        #     self.config[key] = value


class iEDAParameter(AbstractParameter):

    def __init__(self, filename="./config/iEDA/default.json", step=FlowStep.place):
        self.param_path = os.path.join(os.path.dirname(__file__), filename)
        super().__init__(filename, step)
        # print(self._search_space, "search_space iEDA")

    def formatSearchSpace(self, step):
        search_path = os.path.join(os.path.dirname(__file__), "config/iEDA/search_space.json")
        with open(search_path, "r") as f:
            self.params_dict = json.load(f)
            if step != FlowStep.full_flow:
                self._search_space = self.params_dict.get(step.value.lower(), {})
            else:
                self._search_space = dict()
                for k,v in self.params_dict.items():
                    self._search_space.update(v)
    
    def getCurrUpdateParams(self, step):
        key_step = step.value.lower() # str
        print("next_params:", self.next_params)
        curr_param_keys = self.params_dict.get(key_step, {})
        print("curr_param_keys:", curr_param_keys)
        update_params = { k:v for k,v in self.next_params.items() if k in curr_param_keys }
        print("update_params:", update_params)
        return update_params

    def dumpPlaceFlowConfig(self, filename, config=None, step=FlowStep.place):
        key_step = step.value.lower()
        
        if key_step not in self.config:
            return 
        config = self.config[key_step]
        print("dumpPlaceFlowConfig", config)
        # externals = {key:config[key] for key in config if key != "PL"}
        # logging.info(externals)
        update_params = self.getCurrUpdateParams(step)
        for k in update_params:
            # if k != "PL":
            #     del config[k]
            if k in config["PL"]["GP"]["Wirelength"]:
                config["PL"]["GP"]["Wirelength"][k] = update_params[k]
            if k in config["PL"]["GP"]["Density"]:
                config["PL"]["GP"]["Density"][k] = update_params[k]
            if k in config["PL"]["GP"]["Nesterov"]:
                config["PL"]["GP"]["Nesterov"][k] = update_params[k]
        print("dumpPlaceFlowConfig:", config)
        print("filename:", filename)
        with open(filename, 'w') as f:
            json.dump(config, f, indent=4)

    def dumpCTSFlowConfig(self, filename, config=None, step=FlowStep.cts):
        key_step = step.value.lower()
        if key_step not in self.config:
            return 
        config = self.config[key_step]
        update_params = self.getCurrUpdateParams(step)
        curr_search_space = self.params_dict.get(key_step, {})
        for param in curr_search_space:
            param_type = curr_search_space[param].get("type", None)
            if param_type == "str" and param in update_params:
                config[param] = str(update_params[param])
        with open(filename, 'w') as f:
            json.dump(config, f, indent=4)

    def dumpRouteFlowConfig(self, filename, config=None, step=FlowStep.route):
        pass

    def dumpFullFlowConfig(self, filename, config=None, flow_list=None):
        for flow_step in flow_list:
            self.dumpConfigByFlowStep(filename, flow_step)
        
    def dumpOriginalConfig(self, config_paths=None, flow_list=None, filename=None):
        """
        @brief dump original config parameters to json file
        """
        for flow in flow_list:
            step = flow.step
            filename = config_paths[step.value.lower()]
            print("iEDAParameter: dumpOriginalConfig", step, filename)
            self.dumpConfigByFlowStep(filename, step)

class InnovusParameter(AbstractParameter):

    def __init__(self, filename="./config/Innovus/default.json"):
        super().__init__(filename)

    def formatSearchSpace(self,step):
        search_path = os.path.join(os.path.dirname(__file__), "config/Innovus/search_space.json")
        with open(search_path, "r") as f:
            self._search_space = json.load(f)
    
    def parseParams(self):
        self.args = None
        path = "/home/laixinhua/AiEDA/third_party/resource/innovus/scripts"
        for file_name in os.listdir(path):
            print(file_name)

        
    def dumpOriginalConfig(self, filename, step=None, flow_list=None):
        """
        @brief dump original config parameters to json file
        """
        pass
        

class DREAMPlaceParameter(AbstractParameter):
    def __init__(self, filename="./config/DREAMPlace/params.json"):
        super().__init__(filename)

    def formatSearchSpace(self, step):
        search_path = os.path.join(os.path.dirname(__file__), "config/DREAMPlace/search_space.json")
        with open(search_path, "r") as f:
            self._search_space = json.load(f)

        
    def dumpOriginalConfig(self, filename, step=None, flow_list=None):
        """
        @brief dump original config parameters to json file
        """
        # filter wrong keys
        externals = {key:self.config[key] for key in self.config if key != "PL"}
        logging.info(externals)
        for k in externals:
            # remove wrong keys
            if k != "PL":
                del self.config[k]
        #for k in externals:
            if k in self.config["PL"]["GP"]["Wirelength"]:
                self.config["PL"]["GP"]["Wirelength"][k] = externals[k]
            if k in self.config["PL"]["GP"]["Density"]:
                self.config["PL"]["GP"]["Density"][k] = externals[k]
            if k in self.config["PL"]["GP"]["Nesterov"]:
                self.config["PL"]["GP"]["Nesterov"][k] = externals[k]
        # logging.info(externals)
        #logging.info(externals)
        with open(filename, 'w') as f:
            json.dump(self.config, f, indent=4)

if __name__ == "__main__":
    inn = InnovusParameter()
    inn.parseParams()
    



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : flow_manager.py
@Author : yell
@Desc : Manage flow
'''
from dataclasses import field

from workspace.config.json_flow import FlowParser
from database.enum import FlowStep, TaskOption
from flow.flow_db import DbFlow

class FlowManager():
    """Manage flow"""    
    def __init__(self, flow_path : str):
        self.flow_path = flow_path
        self.task : TaskOption = TaskOption.RUN_EDA
        self.flow_list : list = field(default_factory=list)

        self.current_flow = None
        self.pre_flow = None
        self.init_flow()
        
    def get_flow_path(self):
        """get flow path"""
        return self.flow_path
    
    def init_flow(self, flow_path : str=""):
        """set flow path"""
        #save path 
        if(flow_path != ""):            
            self.flow_path = flow_path
        
        #build flow data      
        parser = FlowParser(self.flow_path)
        flow_db_list = parser.get_db()
        if(flow_db_list != None):  
            self.task = flow_db_list['task']
            self.flow_list = flow_db_list['list']
            
    def set_current_flow(self, flow : DbFlow):
        """set current flow"""
        self.current_flow = flow 
        
    def get_current_flow(self):
        """get current flow"""
        return self.current_flow
    
    def set_pre_flow(self, flow : DbFlow):
        """set as pre flow"""
        self.pre_flow = flow
        
    def get_pre_flow(self):
        """get pre flow"""
        return self.pre_flow
    
    def save_flow_state(self, db_flow : DbFlow):
        """save flow state"""
        parser = FlowParser(self.flow_path)
        parser.set_flow_state(db_flow)
    
    def reset_flow_state(self):
        """reset flow state to unstart"""
        parser = FlowParser(self.flow_path)
        parser.reset_flow_state()

    def is_run_eda(self):
        """ if run eda """
        if (self.task == TaskOption.RUN_EDA):
            return True
        else :
            return False
    
    def is_run_fpt(self):
        """run fpt"""
        if (self.task == TaskOption.RUN_FPT):
            return True
        else :
            return False
    
    def set_task(self, task :str):
        """set task"""
        self.task = TaskOption(task)
        
    def get_new_flow(self):
        """get new flow in list"""
        for flow in self.flow_list :
            if flow.is_finish() is not True:
                return flow
            
        return None
    
    def find_flow(self, flow_step : FlowStep):
        for flow in self.flow_list:
            if flow_step == flow.step :
                return flow



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : enum.py
@Author : yell
@Desc : enum definition
'''

from enum import Enum

class TaskState(Enum):
    """task running state
    """
    Unstart = "unstart"
    Success = "success"
    Ongoing = "ongoing"
    Imcomplete = "incomplete"

class TaskOption(Enum):
    """task job
    """
    NO_TASK = None
    RUN_EDA = "run_eda"
    RUN_FPT = "run_fpt"
    RUN_SRCX = "run_starrc" 

class EdaTool(Enum):
    """EDA tool for different step.
    """
    EDI = "edi"
    IEDA = "iEDA"
    DC = "DC"
    INNOVUS = "innovus"
    PT = "PT"
    CALIBRE = "calibre"
    DREAMPLACE = "dreamplace"
    XPLACE = "Xplace"
    OPENROAD = "OpenROAD"
    STARRC = "StarRC"
        
class Process(Enum):
    """chip process
    """
    process_110nm = 1
    process_t28 = 2
    process_sky130 = 3
    
class FlowStep(Enum):
    """PR step
    """
    NoStep = ""
    initFlow = "initFlow"
    initDB = "initDB"
    edi = "edi"
    floorplan = "floorplan"
    pdn = "PDN"
    prePlace = "prePlace"
    place = "place"
    globalPlace = "gp"
    detailPlace = "dp"
    cts = "CTS"
    route = "route"
    globalRouting = "gr"
    detailRouting = "dr"
    eco = "eco"
    fixFanout = "fixFanout"
    optDrv = "optDrv"
    optHold = "optHold"
    optSetup = "optSetup"
    legalization = "legalization"
    filler = "filler"
    drc = "drc"
    sta = "sta"
    rcx = "rcx"
    gds = "gds"
    full_flow = 'full_flow'
    
class FptState(Enum):
    """FPT State
    """
    off = "off"
    on = "on"
    
class PTState(Enum):
    """PT
    """
    off = "off"
    on = "on"

class TrackDirection(Enum):
    """track direction
    """
    none = ""
    horizontal = "H"
    vertical = "V"
    
class LayerType(Enum):
    """layer type
    """
    none = ""
    routing = "routing"
    cut = "cut"
    
class CellType(Enum):
    """cell type
    """
    none = ""
    pad = "pad"
    core = "core"
    
class OrientType(Enum):
    """cell type
    """
    none = ""
    N_R0 = "N"
    W_R90 = "W"
    S_R180 = "S"
    E_R270 = "E"
    FN_MY = "FN"
    FE_MY90 = "FE"
    FS_MX = "FS"
    FW_MX90 = "FW"
    
class PlaceStatus(Enum):
    """placement status
    """
    none = ""
    fixed = "fixed"
    cover = "cover"
    placed = "placed"
    unplaced = "unplaced"
    
    
class NetType(Enum):
    """net type
    """
    none = ""
    signal = "signal"
    clock = "clock"
    power = "power"
    ground = "ground"
    
class EvalCongestionType(Enum):
    """congestion type in evaluation
    """
    none = 0
    instance_density = 1
    pin_density = 2
    net_congestion = 3
    gr_congestion = 4
    macro_margin_h = 5
    macro_margin_v = 6
    continuous_white_space = 7
    macro_margin = 8
    macro_channel = 9
    
class EvalRudyType(Enum):
    """RUDY type in evaluation
    """
    none = 0
    rudy = 1
    pin_rudy = 2
    lut_rudy = 3
    
class EvalInstanceStatus(Enum):
    """congestion type in evaluation
    """
    none = 0
    fixed = 1
    cover = 2
    placed = 3
    unplaced = 4

class EvalWirelengthType(Enum):
    """wirelength type in evaluation
    """
    none = 0
    hpwl = 1
    flute = 2
    b2b = 3
    egr = 4

class EvalDirection(Enum):
    """direction type in evaluation
    """
    none = 0
    h = 1
    v = 2
    
class FeatureOption(Enum):
    """feature options
    """
    NoFeature = None
    summary = "summary" # default
    tools = "tool"
    eval_map = "eval_map"
    drc = "drc"
    eval = "eval"
    timing_eval = "timing_eval"
    baseline_drc = "baseline_drc"
    baseline_sta = "baseline_sta"
    baseline_power = "baseline_power"
    large_model = "large_model"

class DSEMethod(Enum):
    """different EDA DSE method.
    """
    WANDB = "wandb"
    OPTUNA = "optuna"
    NNI = "nni"



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : json_parser.py
@Author : yell
@Desc : json parser 
'''

import json
import gzip
import os
import logging


class JsonParser:
    """basic json parser"""

    def __init__(self, json_path: str):
        self.json_path = json_path
        self.json_data = None

    def set_json_data(self, value: dict):
        self.json_data = value

    def get_json_data(self):
        ''' access to json data '''
        return self.json_data

    def get_value(self, node_json, key):
        if key in node_json:
            return node_json[key]
        else:
            return None

    def read(self):
        ''' Json reader '''
        if not os.path.exists(self.json_path):
            logging.info("error : json file not exist. path = %s",
                         self.json_path)
            return False
        else:
            pass
            # logging.info("info : start parsing file : %s", self.json_path)

        try:
            # compressed
            if self.json_path.endswith(".gz"):
                with gzip.open(self.json_path, 'r') as f:
                    unzip_data = f.read().decode('utf-8')
                    self.json_data = json.loads(unzip_data)
                    # logging.info("info : parsing file sucess")
                    return True
            else:
                with open(self.json_path, 'r', encoding='utf-8') as f_reader:
                    self.json_data = json.load(f_reader)
                    # logging.info("info : parsing file sucess")
                    return True
        except json.JSONDecodeError:
            logging.error(
                "error : json file format error. path = %s", self.json_path)
            return False

    def read_create(self):
        if not os.path.exists(self.json_path):
            # create file
            self.json_data = {}
            self.write()

        # with open(self.json_path, 'r', encoding='utf-8') as f_reader:
        #     self.json_data = json.load(f_reader)

        # logging.info("info : parsing file sucess")
        return self.read()

    def write(self, dict_value: dict = None):
        if dict_value != None:
            self.json_data = dict_value

        ''' Json writer '''
        if self.json_path.endswith(".gz"):
            with gzip.open(self.json_path, 'wb') as f:
                zip_data = json.dumps(self.json_data, indent=4)
                f.write(zip_data.encode('utf-8'))
                logging.info("info : write file sucess")
        else:
            with open(self.json_path, 'w', encoding='utf-8') as f_writer:
                json.dump(self.json_data, f_writer, indent=4)
                # json.dump(self.json_data, f_writer)
                # print(self.json_data)
                logging.info("info : write file sucess")

        return True

    def get_value(self, dict_node, key):
        if isinstance(key, str) and key in dict_node:
            return dict_node[key]

        if isinstance(key, list):
            dict_word = dict_node
            for word in key:
                if word in dict_word:
                    dict_word = dict_word[word]
                else:
                    return None

            return dict_word

        return None



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
iEDA流程运行模块

提供命令行接口和编程接口来运行iEDA EDA流程
"""

import sys
import os
import json
from typing import List, Optional
import argparse
from pathlib import Path

# 库模式导入 - 使用相对导入
try:
    # 库模式：从已安装的包导入
    from aieda.utility.folder_permission import FolderPermissionManager
    from aieda.database.enum import EdaTool, FlowStep
    from aieda.flow.flow_db import DbFlow
    from aieda.engine.cts import EngineCTS
    from aieda.engine.placement import EnginePlacement
    from aieda.engine.routing import EngineRouting
    from aieda.engine.net_opt import EngineNetOpt 
    from aieda.engine.timing_opt import EngineTimingOpt
    from aieda.engine.gds import EngineGDS
    from aieda.flow.run_flow_ieda import RunFlowIEDA
except ImportError:
    # 开发模式：使用相对导入
    try:
        from ...utility.folder_permission import FolderPermissionManager
        from ...database.enum import EdaTool, FlowStep
        from ...flow.flow_db import DbFlow
        from ...engine.cts import EngineCTS
        from ...engine.placement import EnginePlacement
        from ...engine.routing import EngineRouting
        from ...engine.net_opt import EngineNetOpt 
        from ...engine.timing_opt import EngineTimingOpt
        from ...engine.gds import EngineGDS
        from ...flow.run_flow_ieda import RunFlowIEDA
    except ImportError:
        # 兼容原始的绝对路径导入方式
        current_dir = os.path.split(os.path.abspath(__file__))[0]
        aieda_dir = current_dir.rsplit('/', 2)[0]
        sys.path.append(aieda_dir)
        
        from utility.folder_permission import FolderPermissionManager
        from database.enum import EdaTool, FlowStep
        from flow.flow_db import DbFlow
        from engine.cts import EngineCTS
        from engine.placement import EnginePlacement
        from engine.routing import EngineRouting
        from engine.net_opt import EngineNetOpt 
        from engine.timing_opt import EngineTimingOpt
        from engine.gds import EngineGDS
        from flow.run_flow_ieda import RunFlowIEDA


class FlowManager:
    """EDA流程管理器"""
    
    def __init__(self, design: str, workspace_path: str, eda_tool: str = "iEDA"):
        """初始化流程管理器
        
        Args:
            design: 设计名称
            workspace_path: 工作空间路径模板，包含{design}占位符
            eda_tool: EDA工具类型
        """
        self.design = design
        self.workspace_path = workspace_path
        self.eda_tool = eda_tool
        
        # 设置环境变量
        os.environ['eda_tool'] = eda_tool
        
        # 构建实际工作空间路径
        self.dir_workspace = workspace_path.format(design=design)
        
        # 创建流程运行器
        self.flow_runner = RunFlowIEDA(dir_workspace=self.dir_workspace)
    
    def reset_flow_steps(self, step: Optional[str] = None) -> None:
        """重置流程步骤状态"""
        reset_flow_steps(self.design, self.workspace_path, step)
    
    def run_flow(self) -> bool:
        """运行完整流程"""
        try:
            result = self.flow_runner.run()
            return result
        except Exception as e:
            print(f"流程运行失败: {e}")
            return False
    
    def get_flow_status(self) -> dict:
        """获取流程状态"""
        flow_json_path = os.path.join(self.dir_workspace, "config", "flow.json")
        
        if not os.path.exists(flow_json_path):
            return {"error": f"流程配置文件不存在: {flow_json_path}"}
        
        try:
            with open(flow_json_path, 'r', encoding='utf-8') as f:
                flow_data = json.load(f)
            return flow_data
        except Exception as e:
            return {"error": f"读取流程配置失败: {e}"}


def reset_flow_steps(design: str, workspace_path: str, step: Optional[str] = None) -> None:
    """
    修改 flow.json 文件，根据指定的步骤调整流程状态：
    - 将指定步骤及其之后的所有步骤状态设置为 "unstart"
    - 将指定步骤之前的所有步骤状态设置为 "success"
    
    参数:
        design: 设计名称
        workspace_path: 工作空间路径，包含{design}占位符
        step: 从哪个步骤开始执行，如果为 None 则重置所有步骤为 "unstart"
    
    返回:
        None
    """
    # 构建 flow.json 文件路径
    dir_workspace = workspace_path.format(design=design)
    flow_json_path = os.path.join(dir_workspace, "config", "flow.json")
    
    # 检查文件是否存在
    if not os.path.exists(flow_json_path):
        print(f"错误: 文件不存在 - {flow_json_path}")
        return
    
    try:
        # 读取 flow.json 文件
        with open(flow_json_path, 'r', encoding='utf-8') as f:
            flow_data = json.load(f)
        
        # 修改状态
        found_step = False
        success_count = 0
        unstart_count = 0
        
        for item in flow_data.get("flow", []):
            current_step = item.get("step")
            current_state = item.get("state")
            
            # 如果找到指定的起始步骤，标记为已找到
            if step and current_step == step:
                found_step = True
            
            # 如果没有指定步骤，或者已经找到指定步骤，则设置为 "unstart"
            if step is None or found_step:
                if current_state != "unstart":
                    item["state"] = "unstart"
                    unstart_count += 1
            # 如果在指定步骤之前，则设置为 "success"
            elif step and not found_step:
                if current_state != "success":
                    item["state"] = "success"
                    success_count += 1
        
        # 写入修改后的数据
        with open(flow_json_path, 'w', encoding='utf-8') as f:
            json.dump(flow_data, f, indent=4)
        
        # 打印修改结果
        if step is None:
            print(f"已将所有 {unstart_count} 个步骤的状态重置为 'unstart'")
        elif not found_step:
            print(f"错误: 未找到步骤 '{step}'")
        else:
            print(f"已将 '{step}' 之前的 {success_count} 个步骤设置为 'success'")
            print(f"已将 '{step}' 及其之后的 {unstart_count} 个步骤设置为 'unstart'")
            
    except json.JSONDecodeError:
        print(f"错误: 无法解析 JSON 文件 - {flow_json_path}")
    except Exception as e:
        print(f"错误: {str(e)}")


def main():
    """命令行入口函数"""
    # 创建参数解析器
    parser = argparse.ArgumentParser(description='iEDA流程运行工具')
    
    # 添加参数
    parser.add_argument('--design', type=str, default='gcd', 
                        help='设计名称, 默认值为gcd')
    parser.add_argument('--path', type=str, 
                        default='/data/project_share/yhqiu/{design}/workspace', 
                        help='工作空间路径，需包含{design}占位符')
    parser.add_argument('--step', type=str, default=None, 
                        help='从哪个阶段开始执行流程, 默认从头开始执行全流程。'
                             '可选: fixFanout, place, CTS, optDrv, optHold, legalization, route, filler')
    parser.add_argument('--tool', type=str, default='iEDA',
                        help='EDA工具类型，默认为iEDA')
    parser.add_argument('--status', action='store_true',
                        help='只查看流程状态，不运行流程')
    
    # 解析命令行参数
    args = parser.parse_args()
    
    # 创建流程管理器
    flow_manager = FlowManager(
        design=args.design,
        workspace_path=args.path,
        eda_tool=args.tool
    )
    
    # 如果只是查看状态
    if args.status:
        status = flow_manager.get_flow_status()
        print(json.dumps(status, indent=2, ensure_ascii=False))
        return
    
    # 重置流程步骤
    flow_manager.reset_flow_steps(step=args.step)
    
    # 运行流程
    print(f"开始运行 {args.tool} 流程...")
    print(f"设计: {args.design}")
    print(f"工作空间: {flow_manager.dir_workspace}")
    if args.step:
        print(f"从步骤开始: {args.step}")
    
    success = flow_manager.run_flow()
    
    if success:
        print("✅ 流程运行成功!")
        sys.exit(0)
    else:
        print("❌ 流程运行失败!")
        sys.exit(1)


# 编程接口函数
def run_ieda_flow(design: str, workspace_path: str, step: Optional[str] = None, 
                  eda_tool: str = "iEDA") -> bool:
    """编程接口：运行iEDA流程
    
    Args:
        design: 设计名称
        workspace_path: 工作空间路径模板
        step: 从哪个步骤开始，None表示从头开始
        eda_tool: EDA工具类型
    
    Returns:
        bool: 是否成功
    """
    flow_manager = FlowManager(design, workspace_path, eda_tool)
    flow_manager.reset_flow_steps(step)
    return flow_manager.run_flow()


if __name__ == "__main__":
    main()


"""EDA引擎模块"""
from .cts import EngineCTS
from .placement import EnginePlacement
from .routing import EngineRouting
from .net_opt import EngineNetOpt
from .timing_opt import EngineTimingOpt
from .gds import EngineGDS